import copy
import numbers
from typing import Any, Dict, List, Optional, Type, Union

from jsonpath_ng.ext import parse

from rubicon_ml.client import Experiment, Project, Rubicon


class RubiconJSON:
    """`RubiconJSON` converts top-level `rubicon_ml` objects, `projects`, and `experiments`
    into a JSON structured dictionary for JSONPath-like querying with `jsonpath-ng`.

    Parameters
    ----------
    rubicon_objects : rubicon.client.Rubicon or `list` of type rubicon.client.Rubicon
        Top-level `rubicon-ml` objects to convert to JSON for querying.
    projects : rubicon.client.Project or `list` of type rubicon.client.Project
        `rubicon-ml` projects to convert to JSON for querying.
    experiments : rubicon.client.Experiment or `list` of type rubicon.client.Experiment
        `rubicon-ml` experiments to convert to JSON for querying.
    """

    def __init__(
        self,
        rubicon_objects: Optional[List[Rubicon]] = None,
        projects: Optional[List[Project]] = None,
        experiments: Optional[List[Experiment]] = None,
    ):
        if rubicon_objects:
            rubicon_objects = self._validate_input(rubicon_objects, Rubicon, "rubicon_objects")
        if projects:
            projects = self._validate_input(projects, Project, "projects")
        if experiments:
            experiments = self._validate_input(experiments, Experiment, "experiments")

        self._json = self._convert_to_json(rubicon_objects, projects, experiments)
        self._json_numeric = self._convert_to_json(
            rubicon_objects, projects, experiments, filter_nonnumeric=True
        )

    def _validate_input(
        self,
        objects: Union[
            Rubicon, List[Rubicon], Project, List[Project], Experiment, List[Experiment]
        ],
        obj_cls: Union[Type[Rubicon], Type[Project], Type[Experiment]],
        obj_name: str,
    ):
        formatted_objects = [objects] if not isinstance(objects, list) else objects

        if not all([isinstance(o, obj_cls) for o in formatted_objects]):
            raise ValueError(f"`{obj_name}` must be `list` of type `{obj_cls}`.")

        return formatted_objects

    def _convert_to_json(
        self,
        rubicon_objects: Optional[List[Rubicon]] = None,
        projects: Optional[List[Project]] = None,
        experiments: Optional[List[Experiment]] = None,
        filter_nonnumeric: bool = False,
    ):
        rubicon_json = {}

        if rubicon_objects is not None:
            rubicon_json["project"] = self._rubicon_to_json(
                rubicon_objects, filter_nonnumeric=filter_nonnumeric
            )["project"]

        if projects is not None:
            project_json = rubicon_json.get("project", [])
            project_json.extend(
                self._projects_to_json(projects, filter_nonnumeric=filter_nonnumeric)["project"]
            )
            rubicon_json["project"] = project_json

        if experiments is not None:
            rubicon_json["experiment"] = self._experiments_to_json(
                experiments, filter_nonnumeric=filter_nonnumeric
            )["experiment"]

        return rubicon_json

    def _experiments_to_json(self, experiments: List[Experiment], filter_nonnumeric: bool = False):
        rubicon_json: Dict[str, Any] = {"experiment": []}

        for e in experiments:
            experiment_json = copy.deepcopy(e._domain.__dict__)
            experiment_json["feature"] = [f._domain.__dict__ for f in e.features()]

            experiment_json["parameter"] = []
            for parameter in e.parameters():
                if not filter_nonnumeric or isinstance(parameter.value, numbers.Number):
                    experiment_json["parameter"].append(parameter._domain.__dict__)

            experiment_json["metric"] = []
            for metric in e.metrics():
                if not filter_nonnumeric or isinstance(metric.value, numbers.Number):
                    experiment_json["metric"].append(metric._domain.__dict__)

            experiment_json["artifact"] = [a._domain.__dict__ for a in e.artifacts()]
            experiment_json["dataframe"] = [d._domain.__dict__ for d in e.dataframes()]

            rubicon_json["experiment"].append(experiment_json)

        return rubicon_json

    def _projects_to_json(self, projects: List[Project], filter_nonnumeric: bool = False):
        rubicon_json: Dict[str, Any] = {"project": []}

        for p in projects:
            project_json = p._domain.__dict__
            project_json["artifact"] = [a._domain.__dict__ for a in p.artifacts()]
            project_json["dataframe"] = [d._domain.__dict__ for d in p.dataframes()]

            experiment_json = self._experiments_to_json(
                p.experiments(), filter_nonnumeric=filter_nonnumeric
            )
            project_json["experiment"] = experiment_json["experiment"]

            rubicon_json["project"].append(project_json)

        return rubicon_json

    def _rubicon_to_json(self, rubicon_objects: List[Rubicon], filter_nonnumeric: bool = False):
        rubicon_json: Dict[str, Any] = {"project": []}

        for r in rubicon_objects:
            rubicon_json["project"].extend(
                self._projects_to_json(r.projects(), filter_nonnumeric=filter_nonnumeric)["project"]
            )

        return rubicon_json

    def search(self, query: str):
        """
        Query the JSON generated from the RubiconJSON instantiation in a JSONPath-like manner.
        Can return queries as `rubicon_ml.client` objects by specifying return_type parameter.
        Will return as JSON structured `dict` by default.

        Parameters
        ----------
        query: JSONPath-like query
        """

        if ">" in query or "<" in query:
            # non-numerics break greater than and less than comparisons in `jsonpath_ng`
            # so we use the json with replaced non-numeric values when '>' or '<' appear
            # in the `query`
            json = self._json_numeric
        else:
            json = self._json

        return parse(query).find(json)

    @property
    def json(self):
        """
        The json representation of the `rubicon-ml` objects.
        """
        return self._json

    @property
    def json_numeric(self):
        """
        The json representation of the `rubicon-ml` objects with numeric values.
        """
        return self._json_numeric



from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

from rubicon_ml.domain.mixin import InitMixin

if TYPE_CHECKING:
    from datetime import datetime

    from rubicon_ml.domain.utils import TrainingMetadata


@dataclass(init=False)
class Project(InitMixin):
    """A domain-level project.

    Parameters
    ----------
    name : str
        The project's name.
    created_at : datetime, optional
        The date and time the project was created. Defaults to `None` and uses
        `datetime.datetime.now` to generate a UTC timestamp. `created_at` should be
        left as `None` to allow for automatic generation.
    description : str, optional
        A description of the project. Defaults to `None`.
    github_url : str, optional
        The URL of the GitHub repository associated with this project. Defaults to
        `None`.
    id : str, optional
        The project's unique identifier. Defaults to `None` and uses `uuid.uuid4`
        to generate a unique ID. `id` should be left as `None` to allow for automatic
        generation.
    training_metadata : rubicon_ml.domain.utils.TrainingMetadata, optional
        Additional metadata pertaining to any data this project was trained on.
        Defaults to `None`.
    """

    name: str
    created_at: Optional["datetime"] = None
    description: Optional[str] = None
    github_url: Optional[str] = None
    id: Optional[str] = None
    training_metadata: Optional["TrainingMetadata"] = None

    def __init__(
        self,
        name: str,
        created_at: Optional["datetime"] = None,
        description: Optional[str] = None,
        github_url: Optional[str] = None,
        id: Optional[str] = None,
        training_metadata: Optional["TrainingMetadata"] = None,
        **kwargs,
    ):
        """Initialize this domain project."""

        self._check_extra_kwargs(kwargs)

        self.name = name
        self.created_at = self._init_created_at(created_at)
        self.description = description
        self.github_url = github_url
        self.id = self._init_id(id)
        self.training_metadata = training_metadata



import os

from rubicon_ml.repository import BaseRepository
from rubicon_ml.repository.utils import json


class LocalRepository(BaseRepository):
    """The local repository persists Rubicon data to a local
    filesystem.

    Parameters
    ----------
    root_dir : str
        Absolute path to the root directory to persist Rubicon
        data to.
    """

    PROTOCOL = "file"

    def _persist_bytes(self, bytes_data, path):
        """Persists the raw bytes `bytes_data` to the local
        path defined by `path`.
        """
        self.filesystem.mkdirs(os.path.dirname(path), exist_ok=True)

        with self.filesystem.open(path, "wb") as f:
            f.write(bytes_data)

    def _persist_domain(self, domain, path):
        """Persists the Rubicon object `domain` to the local
        path defined by `path`.
        """
        json_domain = json.dumps(domain)

        self.filesystem.mkdirs(os.path.dirname(path), exist_ok=True)

        with self.filesystem.open(path, "w") as f:
            f.write(json_domain)



import dataclasses
import json
from base64 import b64decode, b64encode
from datetime import date, datetime

import numpy as np

from rubicon_ml.domain.utils import TrainingMetadata


class DomainJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        """
        Note - if we need to support nested objects
        within dataclasses, we need to leverage asdict()
        """
        if isinstance(obj, datetime):
            return {"_type": "datetime", "value": obj.strftime("%Y-%m-%d %H:%M:%S.%f")}
        elif isinstance(obj, date):
            return {"_type": "date", "value": obj.isoformat()}
        elif isinstance(obj, set):
            return {"_type": "set", "value": list(obj)}
        elif isinstance(obj, TrainingMetadata):
            return {"_type": "training_metadata", "value": obj.training_metadata}
        elif isinstance(obj, (np.generic, np.ndarray)):
            return {
                "_type": "numpy",
                "_dtype": np.lib.format.dtype_to_descr(obj.dtype),
                "_shape": obj.shape,
                "value": b64encode(obj.tobytes()).decode(),
            }
        elif dataclasses.is_dataclass(obj):
            return obj.__dict__
        else:
            return super().default(obj)  # pragma: no cover


class DomainJSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        if obj.get("_type") == "datetime":
            return datetime.strptime(obj.get("value"), "%Y-%m-%d %H:%M:%S.%f")
        elif obj.get("_type") == "date":
            return date.fromisoformat(obj.get("value"))
        elif obj.get("_type") == "set":
            return set(obj.get("value"))
        elif obj.get("_type") == "training_metadata":
            return TrainingMetadata([(*o,) for o in obj.get("value")])
        elif obj.get("_type") == "numpy":
            dtype = np.lib.format.descr_to_dtype(obj.get("_dtype"))
            shape = obj.get("_shape")
            value = np.frombuffer(b64decode(obj.get("value")), dtype)

            return value.reshape(shape) if shape else value[0]
        else:
            return obj


def dump(data, open_file, **kwargs):
    return json.dump(data, open_file, cls=DomainJSONEncoder, **kwargs)


def dumps(data, **kwargs):
    return json.dumps(data, cls=DomainJSONEncoder, **kwargs)


def load(open_file, **kwargs):
    return json.load(open_file, cls=DomainJSONDecoder, **kwargs)


def loads(data, **kwargs):
    return json.loads(data, cls=DomainJSONDecoder, **kwargs)



from dataclasses import dataclass
from typing import TYPE_CHECKING, List, Optional

from rubicon_ml.domain.mixin import CommentMixin, InitMixin, TagMixin

if TYPE_CHECKING:
    from datetime import datetime


@dataclass(init=False)
class Feature(CommentMixin, InitMixin, TagMixin):
    """A domain-level feature.

    name : str
        The feature's name.
    comments : list of str, optional
        Additional text information and observations about the feature. Defaults to
        `None`.
    created_at : datetime, optional
        The date and time the feature was created. Defaults to `None` and uses
        `datetime.datetime.now` to generate a UTC timestamp. `created_at` should be
        left as `None` to allow for automatic generation.
    description : str, optional
        A description of the feature. Defaults to `None`.
    id : str, optional
        The feature's unique identifier. Defaults to `None` and uses `uuid.uuid4`
        to generate a unique ID. `id` should be left as `None` to allow for automatic
        generation.
    importance : float, optional
        The feature's calculated importance value. Defaults to `None`.
    tags : list of str, optional
        The values this feature is tagged with. Defaults to `None`.
    """

    name: str
    comments: Optional[List[str]] = None
    created_at: Optional["datetime"] = None
    description: Optional[str] = None
    id: Optional[str] = None
    importance: Optional[float] = None
    tags: Optional[List[str]] = None

    def __init__(
        self,
        name: str,
        comments: Optional[List[str]] = None,
        created_at: Optional["datetime"] = None,
        description: Optional[str] = None,
        id: Optional[str] = None,
        importance: Optional[float] = None,
        tags: Optional[List[str]] = None,
        **kwargs,
    ):
        """Initialize this domain feature."""

        self._check_extra_kwargs(kwargs)

        self.name = name
        self.comments = comments or []
        self.created_at = self._init_created_at(created_at)
        self.description = description
        self.id = self._init_id(id)
        self.importance = importance
        self.tags = tags or []



"""``schema`` submodule initialization."""

from rubicon_ml.schema.registry import (
    available_schema,
    get_schema,
    get_schema_name,
    register_schema,
)

__all__ = ["available_schema", "get_schema", "get_schema_name", "register_schema"]



from rubicon_ml.viz.dashboard import Dashboard
from rubicon_ml.viz.dataframe_plot import DataframePlot
from rubicon_ml.viz.experiments_table import ExperimentsTable
from rubicon_ml.viz.metric_correlation_plot import MetricCorrelationPlot
from rubicon_ml.viz.metric_lists_comparison import MetricListsComparison

__all__ = [
    "Dashboard",
    "DataframePlot",
    "ExperimentsTable",
    "MetricListsComparison",
    "MetricCorrelationPlot",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Literal, Optional, Union

from rubicon_ml.client import Base, CommentMixin, TagMixin
from rubicon_ml.client.utils.exception_handling import failsafe
from rubicon_ml.exceptions import RubiconException

if TYPE_CHECKING:
    from rubicon_ml.client import Experiment, Project
    from rubicon_ml.domain import Dataframe as DataframeDomain


class Dataframe(Base, TagMixin, CommentMixin):
    """A client dataframe.

    A `dataframe` is a two-dimensional, tabular dataset with
    labeled axes (rows and columns) that provides value to the
    model developer and/or reviewer when visualized.

    For example, confusion matrices, feature importance tables
    and marginal residuals can all be logged as a `dataframe`.

    A `dataframe` is logged to a `project` or an `experiment`.

    Parameters
    ----------
    domain : rubicon.domain.Dataframe
        The dataframe domain model.
    parent : rubicon.client.Project or rubicon.client.Experiment
        The project or experiment that the artifact is
        logged to.
    """

    def __init__(self, domain: DataframeDomain, parent: Union[Experiment, Project]):
        super().__init__(domain, parent._config)

        self._domain: DataframeDomain

        self._data = None
        self._parent = parent

    @failsafe
    def get_data(self, df_type: Literal["pandas", "dask"] = "pandas"):
        """Loads the data associated with this Dataframe
        into a `pandas` or `dask` dataframe.

        Parameters
        ----------
        df_type : str, optional
            The type of dataframe to return. Valid options include
            ["dask", "pandas"]. Defaults to "pandas".
        """
        project_name, experiment_id = self.parent._get_identifiers()
        return_err = None
        for repo in self.repositories:
            try:
                self._data = repo.get_dataframe_data(
                    project_name,
                    self.id,
                    experiment_id=experiment_id,
                    df_type=df_type,
                )
            except Exception as err:
                return_err = err
            else:
                return self._data

        raise RubiconException(return_err)

    @failsafe
    def plot(
        self,
        df_type: Literal["pandas", "dask"] = "pandas",
        plotting_func: Optional[Callable] = None,
        **kwargs,
    ):
        """Render the dataframe using `plotly.express`.

        Parameters
        ----------
        df_type : str, optional
            The type of dataframe. Can be either `pandas` or `dask`.
            Defaults to 'pandas'.
        plotting_func : function, optional
            The `plotly.express` plotting function used to visualize the
            dataframes. Available options can be found at
            https://plotly.com/python-api-reference/plotly.express.html.
            Defaults to `plotly.express.line`.
        kwargs : dict, optional
            Keyword arguments to be passed to `plotting_func`. Available options
            can be found in the documentation of the individual functions at the
            URL above.

        Examples
        --------
        >>> # Log a line plot
        >>> dataframe.plot(x='Year', y='Number of Subscriptions')

        >>> # Log a bar plot
        >>> import plotly.express as px
        >>> dataframe.plot(plotting_func=px.bar, x='Year', y='Number of Subscriptions')
        """
        try:
            import plotly.express as px

            if plotting_func is None:
                plotting_func = px.line
        except ImportError:
            raise RubiconException(
                "`ui` extras are required for plotting. Install with `pip install rubicon-ml[ui]`."
            )

        return plotting_func(self.get_data(df_type=df_type), **kwargs)

    @property
    def id(self):
        """Get the dataframe's id."""
        return self._domain.id

    @property
    def name(self):
        """Get the dataframe's name."""
        return self._domain.name

    @property
    def description(self):
        """Get the dataframe's description."""
        return self._domain.description

    @property
    def created_at(self):
        """Get the time this dataframe was created."""
        return self._domain.created_at

    @property
    def parent(self):
        """Get the dataframe's parent client object."""
        return self._parent



from typing import TYPE_CHECKING, Optional, Union

import fsspec
import yaml

if TYPE_CHECKING:
    from rubicon_ml.viz import DataframePlot
    from rubicon_ml.viz.experiments_table import ExperimentsTable
    from rubicon_ml.viz.metric_correlation_plot import MetricCorrelationPlot
    from rubicon_ml.viz.metric_lists_comparison import MetricListsComparison


def publish(
    experiments,
    visualization_object: Optional[
        Union["ExperimentsTable", "MetricCorrelationPlot", "DataframePlot", "MetricListsComparison"]
    ] = None,
    output_filepath=None,
    base_catalog_filepath=None,
):
    """Publish experiments to an `intake` catalog that can be
    read by the `intake-rubicon` driver.

    Parameters
    ----------
    experiments : list of rubicon_ml.client.experiment.Experiment
        The experiments to publish.
    output_filepath : str, optional
        The absolute or relative local filepath or S3 bucket
        and key to log the generated YAML file to. S3 buckets
        must be prepended with 's3://'. Defaults to None,
        which disables writing the generated YAML.
    base_catalog_filepath : str, optional
        Similar to `output_filepath` except this argument is used as a
        base base file to update an existing intake catalog. Defaults to None,
        creating a new intake catalog.

    Returns
    -------
    str
        The YAML string representation of the `intake` catalog
        containing the experiments `experiments`.
    """

    if base_catalog_filepath is not None:
        return _update_catalog(
            base_catalog_filepath=base_catalog_filepath,
            new_experiments=experiments,
            # pass to update catalog
            new_visualization=visualization_object,
            output_filepath=output_filepath,
        )
    # if new file then just pass viz object straight to build catalog
    catalog = _build_catalog(experiments=experiments, visualization=visualization_object)
    catalog_yaml = yaml.dump(catalog)

    if output_filepath is not None:
        with fsspec.open(output_filepath, "w", auto_mkdir=False) as f:
            f.write(catalog_yaml)

    return catalog_yaml


def _update_catalog(
    base_catalog_filepath, new_experiments, new_visualization, output_filepath=None
):
    """Helper function to update exisiting intake catalog.

    Parameters
    ----------
    base_catalog_filepath : str
        the absolute or relative catalog filepath or S3 bucket
        and key to log the generated YAML file to. S3 buckets
        must be prepended with 's3://. Retrieved from the parameter
        of the publish function. NOT optional
    new_experiments : list of rubicon_ml.client.experiment.Experiment
         The new experiments to append to the catalog at
        `base_catalog_filepath`.
    output_catalog_filepath : str, optional
        absolute or relative filepath or S3 bucket
        and key to log the generated YAML file to. (S3 buckets
        must be prepended with 's3://) to  output the updated catalog into.
        Default is None, which resolves to dumping updated catalog into
        base_catalog_filepath path (primary use-case)

    Returns
    -------
    dict
        The dictionary of all sources given as experiments to eventually publish
    """
    # rebuild a temp catalog with new visualization
    updated_catalog = _build_catalog(experiments=new_experiments, visualization=new_visualization)

    with fsspec.open(base_catalog_filepath, "r") as yamlfile:
        curr_catalog = yaml.safe_load(yamlfile)

        curr_catalog["sources"].update(updated_catalog["sources"])

    resulting_filepath = base_catalog_filepath if not output_filepath else output_filepath

    with fsspec.open(resulting_filepath, "w") as yamlfile:
        yaml.safe_dump(curr_catalog, yamlfile)
        updated_catalog = yaml.dump(curr_catalog)

    return updated_catalog


def _build_catalog(experiments, visualization):
    from rubicon_ml.viz import DataframePlot
    from rubicon_ml.viz.experiments_table import ExperimentsTable
    from rubicon_ml.viz.metric_correlation_plot import MetricCorrelationPlot
    from rubicon_ml.viz.metric_lists_comparison import MetricListsComparison

    """Helper function to build catalog dictionary from given experiments.

    Parameters
    ----------
    experiments : list of rubicon_ml.client.experiment.Experiment
        The expriments that are used to build the catalog to eventually publish
    Returns
    -------
    str
        The YAML string representation of the `intake` catalog
        containing the experiments `experiments`.
    """

    catalog = {"sources": {}}

    for experiment in experiments:
        appended_experiment_catalog = {
            "driver": "rubicon_ml_experiment",
            "args": {
                "experiment_id": experiment.id,
                "project_name": experiment.project.name,
                "urlpath": experiment.repository.root_dir,
            },
        }

        experiment_catalog_name = f"experiment_{experiment.id.replace('-', '_')}"
        catalog["sources"][experiment_catalog_name] = appended_experiment_catalog

    # create visualization entry to the catalog file
    # visualization is an ExperimentsTable
    if visualization is not None:
        if isinstance(visualization, ExperimentsTable):
            appended_visualization_catalog = {
                "driver": "rubicon_ml_experiment_table",
                "args": {
                    "is_selectable": visualization.is_selectable,
                    "metric_names": visualization.metric_names,
                    "metric_query_tags": visualization.metric_query_tags,
                    "metric_query_type": visualization.metric_query_type,
                    "parameter_names": visualization.parameter_names,
                    "parameter_query_tags": visualization.parameter_query_tags,
                    "parameter_query_type": visualization.parameter_query_type,
                },
            }
            catalog["sources"]["experiment_table"] = appended_visualization_catalog

        # vizualization is a MetricCorrelationPlot
        if isinstance(visualization, MetricCorrelationPlot):
            appended_visualization_catalog = {
                "driver": "rubicon_ml_metric_correlation_plot",
                "args": {
                    "metric_names": visualization.metric_names,
                    "parameter_names": visualization.parameter_names,
                    "selected_metric": visualization.selected_metric,
                },
            }
            catalog["sources"]["metric_correlation_plot"] = appended_visualization_catalog

        # vizualization is a DataframePlot
        if isinstance(visualization, DataframePlot):
            appended_visualization_catalog = {
                "driver": "rubicon_ml_dataframe_plot",
                "args": {
                    "dataframe_name": visualization.dataframe_name,
                    "x": visualization.x,
                    "y": visualization.y,
                },
            }

            # append visualization object to end of catalog file
            catalog["sources"]["dataframe_plot"] = appended_visualization_catalog

        # append visualization object to end of catalog file

        # vizualization is a MetricListsComparison
        if isinstance(visualization, MetricListsComparison):
            appended_visualization_catalog = {
                "driver": "rubicon_ml_metric_list",
                "args": {
                    "column_names": visualization.column_names,
                    "selected_metric": visualization.selected_metric,
                },
            }

            # append visualization object to end of catalog file
            catalog["sources"]["metric_list"] = appended_visualization_catalog

        # append visualization object to end of catalog file

    return catalog



"""Mehtods for interacting with the existing rubicon-ml ``schema``."""

import os
from typing import Any, List

import yaml

RUBICON_SCHEMA_REGISTRY = {
    "h2o__H2OGeneralizedLinearEstimator": lambda: _load_schema(
        os.path.join("schema", "h2o__H2OGeneralizedLinearEstimator.yaml")
    ),
    "h2o__H2OGradientBoostingEstimator": lambda: _load_schema(
        os.path.join("schema", "h2o__H2OGradientBoostingEstimator.yaml")
    ),
    "h2o__H2ORandomForestEstimator": lambda: _load_schema(
        os.path.join("schema", "h2o__H2ORandomForestEstimator.yaml")
    ),
    "h2o__H2OTargetEncoderEstimator": lambda: _load_schema(
        os.path.join("schema", "h2o__H2OTargetEncoderEstimator.yaml")
    ),
    "h2o__H2OXGBoostEstimator": lambda: _load_schema(
        os.path.join("schema", "h2o__H2OXGBoostEstimator.yaml")
    ),
    "lightgbm__LGBMModel": lambda: _load_schema(os.path.join("schema", "lightgbm__LGBMModel.yaml")),
    "lightgbm__LGBMClassifier": lambda: _load_schema(
        os.path.join("schema", "lightgbm__LGBMClassifier.yaml")
    ),
    "lightgbm__LGBMRegressor": lambda: _load_schema(
        os.path.join("schema", "lightgbm__LGBMRegressor.yaml")
    ),
    "sklearn__RandomForestClassifier": lambda: _load_schema(
        os.path.join("schema", "sklearn__RandomForestClassifier.yaml")
    ),
    "xgboost__XGBModel": lambda: _load_schema(os.path.join("schema", "xgboost__XGBModel.yaml")),
    "xgboost__XGBClassifier": lambda: _load_schema(
        os.path.join("schema", "xgboost__XGBClassifier.yaml")
    ),
    "xgboost__XGBRegressor": lambda: _load_schema(
        os.path.join("schema", "xgboost__XGBRegressor.yaml")
    ),
    "xgboost__DaskXGBClassifier": lambda: _load_schema(
        os.path.join("schema", "xgboost__DaskXGBClassifier.yaml")
    ),
    "xgboost__DaskXGBRegressor": lambda: _load_schema(
        os.path.join("schema", "xgboost__DaskXGBRegressor.yaml")
    ),
}


def _load_schema(path: str) -> Any:
    """Loads a schema YAML file from ``path`` relative to this file."""

    full_path = os.path.join(os.path.dirname(__file__), path)
    with open(full_path, "r") as file:
        schema = yaml.safe_load(file)

    return schema


def available_schema() -> List[str]:
    """Get the names of all available schema."""

    return list(RUBICON_SCHEMA_REGISTRY.keys())


def get_schema(name: str) -> Any:
    """Get the schema with name ``name``."""

    if name not in RUBICON_SCHEMA_REGISTRY:
        raise ValueError(
            f"'{name}' is not the name of an available rubicon schema. "
            "For a list of schema names, use `registry.available_schema()`."
        )

    return RUBICON_SCHEMA_REGISTRY[name]()


def get_schema_name(obj: Any) -> str:
    """Get the name of the schema that represents object ``obj``."""

    obj_cls = obj.__class__

    cls_name = obj_cls.__name__
    module_name = obj_cls.__module__.split(".")[0]

    return f"{module_name}__{cls_name}"


def register_schema(name: str, schema: dict):
    """Add a schema to the schema registry."""

    RUBICON_SCHEMA_REGISTRY[name] = lambda: schema



from __future__ import annotations

import contextlib
import json
import os
import pickle
import tempfile
import warnings
import zipfile
from typing import TYPE_CHECKING, Literal, Optional

import fsspec

from rubicon_ml.client.base import Base
from rubicon_ml.client.mixin import CommentMixin, TagMixin
from rubicon_ml.client.utils.exception_handling import failsafe

if TYPE_CHECKING:
    from rubicon_ml.client import Project
    from rubicon_ml.domain import Artifact as ArtifactDomain


class Artifact(Base, TagMixin, CommentMixin):
    """A client artifact.

    An `artifact` is a catch-all for any other type of
    data that can be logged to a file.

    For example, a snapshot of a trained model (.pkl) can
    be logged to the `experiment` created during its run.
    Or, a base model for the model in development can be
    logged to a `project` when leveraging transfer learning.

    An `artifact` is logged to a `project` or an `experiment`.

    Parameters
    ----------
    domain : rubicon.domain.Artifact
        The artifact domain model.
    parent : rubicon.client.Project or rubicon.client.Experiment
        The project or experiment that the artifact is
        logged to.
    """

    def __init__(self, domain: ArtifactDomain, parent: Project):
        super().__init__(domain, parent._config)

        self._data = None
        self._parent = parent

    def _get_data(self):
        """Loads the data associated with this artifact."""
        project_name, experiment_id = self.parent._get_identifiers()
        return_err = None

        self._data = None

        for repo in self.repositories or []:
            try:
                self._data = repo.get_artifact_data(
                    project_name, self.id, experiment_id=experiment_id
                )
            except Exception as err:
                return_err = err
            else:
                return

        if self._data is None:
            self._raise_rubicon_exception(return_err)

    @failsafe
    def get_data(
        self,
        deserialize: Optional[Literal["h2o", "h2o_binary", "h2o_mojo", "pickle", "xgboost"]] = None,
        unpickle: bool = False,  # TODO: deprecate & move to `deserialize`
    ):
        """Loads the data associated with this artifact and
        unpickles if needed.

        Parameters
        ----------
        deseralize : str, optional
            Method to use to deseralize this artifact's data.
            * None to disable deseralization and return the raw data.
            * "h2o" or "h2o_binary" to use `h2o.load_model` to load the data.
            * "h2o_mojo" to use `h2o.import_mojo` to load the data.
            * "pickle" to use pickles to load the data.
            * "xgboost" to use xgboost's JSON loader to load the data as a fitted model.
            Defaults to None.
        unpickle : bool, optional
            Flag indicating whether or not to unpickle artifact data.
            `deserialize` takes precedence. Defaults to False.
            **Deprecated**: Please use `deserialize="pickle"` in the future.
        """
        project_name, experiment_id = self.parent._get_identifiers()
        return_err = None

        if unpickle:
            warnings.warn(
                "`unpickle` is deprecated, please use `deserialize='pickle'` instead",
                DeprecationWarning,
            )
            deserialize = "pickle"

        if deserialize == "h2o":
            warnings.warn(
                "'deserialize' method 'h2o' will be deprecated in a future release,"
                " please use 'h2o_binary' instead.",
                DeprecationWarning,
            )

        for repo in self.repositories or []:
            try:
                if deserialize == "xgboost":
                    # xgboost can only handle string file name locations
                    import xgboost

                    artifact_data_path = repo._get_artifact_data_path(
                        project_name, experiment_id, self.id
                    )
                    data = xgboost.Booster()
                    data.load_model(artifact_data_path)
                else:
                    data = repo.get_artifact_data(
                        project_name, self.id, experiment_id=experiment_id
                    )
            except Exception as err:
                return_err = err
            else:
                if deserialize in [
                    "h2o",
                    "h2o_binary",
                ]:  # "h2o" will be deprecated in a future release
                    import h2o

                    data = h2o.load_model(
                        repo._get_artifact_data_path(project_name, experiment_id, self.id)
                    )
                elif deserialize == "h2o_mojo":
                    import h2o

                    data = h2o.import_mojo(
                        repo._get_artifact_data_path(project_name, experiment_id, self.id)
                    )
                elif deserialize == "pickle":
                    data = pickle.loads(data)

                return data

        self._raise_rubicon_exception(return_err)

    @failsafe
    def get_json(self):
        return json.loads(self.get_data())

    @failsafe
    def download(
        self,
        location: Optional[str] = None,
        name: Optional[str] = None,
        unzip: bool = False,
    ):
        """Download this artifact's data.

        Parameters
        ----------
        location : str, optional
            The absolute or relative local directory or S3
            bucket to download the artifact to. S3 buckets
            must be prepended with 's3://'. Defaults to the
            current local working directory.
        name : str, optional
            The name to give the downloaded artifact file.
            Defaults to the artifact's given name when logged.
        unzip : bool, optional
            True to unzip the artifact data. False otherwise.
            Defaults to False.
        """
        if location is None:
            location = os.getcwd()

        if name is None:
            name = self._domain.name

        if unzip:
            temp_file_context = tempfile.TemporaryDirectory
        else:
            temp_file_context = contextlib.nullcontext

        with temp_file_context() as temp_dir:
            if unzip:
                location_path = os.path.join(temp_dir, "temp_file.zip")
            else:
                location_path = os.path.join(location, name)

            with fsspec.open(location_path, "wb", auto_mkdir=False) as f:
                f.write(self.data)

            if unzip:
                with zipfile.ZipFile(location_path, "r") as zip_file:
                    zip_file.extractall(location)

    @contextlib.contextmanager
    @failsafe
    def temporary_download(self, unzip: bool = False):
        """Temporarily download this artifact's data within a context manager.

        Parameters
        ----------
        unzip : bool, optional
            True to unzip the artifact data. False otherwise.
            Defaults to False.

        Yields
        ------
        file
            An open file pointer into the directory the artifact data was
            temporarily downloaded into. If the artifact is a single file,
            its name is stored in the `artifact.name` attribute.
        """
        with tempfile.TemporaryDirectory() as temp_dir:
            self.download(location=temp_dir, unzip=unzip)

            yield temp_dir

    @property
    def id(self) -> str:
        """Get the artifact's id."""
        return self._domain.id

    @property
    def name(self) -> str:
        """Get the artifact's name."""
        return self._domain.name

    @property
    def description(self) -> str:
        """Get the artifact's description."""
        return self._domain.description

    @property
    def created_at(self):
        """Get the time this dataframe was created."""
        return self._domain.created_at

    @property
    def data(self):
        """Get the artifact's raw data."""
        warnings.warn(
            "`data` is deprecated, use `get_data()` instead",
            DeprecationWarning,
        )
        if self._data is None:
            self._get_data()

        return self._data

    @property
    def parent(self):
        """Get the artifact's parent client object."""
        return self._parent



import fsspec

from rubicon_ml.repository import BaseRepository
from rubicon_ml.repository.utils import json


class S3Repository(BaseRepository):
    """The S3 repository persists Rubicon data to a remote
    S3 bucket.

    S3 credentials can be specified via environment variables
    or the credentials file in '~/.aws'.

    Parameters
    ----------
    root_dir : str
        The full S3 path (including 's3://') to persist
        Rubicon data to.
    """

    PROTOCOL = "s3"

    def __init__(self, root_dir: str, **storage_options):
        self._df_storage_options = storage_options

        self.filesystem = fsspec.filesystem(self.PROTOCOL, **storage_options)
        self.root_dir = root_dir.rstrip("/")

    def _persist_bytes(self, bytes_data, path):
        """Persists the raw bytes `bytes_data` to the S3
        bucket defined by `path`.
        """
        with self.filesystem.open(path, "wb") as f:
            f.write(bytes_data)

    def _persist_domain(self, domain, path):
        """Persists the Rubicon object `domain` to the S3
        bucket defined by `path`.
        """
        json_domain = json.dumps(domain)

        with self.filesystem.open(path, "w") as f:
            f.write(json_domain)



import os
import pprint
import sys
import warnings

import click

from rubicon_ml import Rubicon
from rubicon_ml.client.rubicon_json import RubiconJSON
from rubicon_ml.viz import Dashboard


@click.group()
@click.version_option()
def cli():
    pass


# Top level CLI commands


@cli.command(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@click.option(
    "--root-dir",
    type=click.STRING,
    help="The absolute path to the top level folder holding the rubicon-ml project.",
    required=True,
)
@click.option(  # DEPRECATED
    "--page-size",
    "-ps",
    type=click.IntRange(min=1),
    help="The number of rows that will be displayed on a page within the experiment table.",
    default=10,
)
@click.option(
    "--project-name",
    "-pn",
    type=click.STRING,
    help="The name of the rubicon-ml project to visualize.",
    default=None,
)
@click.option(
    "--host", "-h", type=click.STRING, help="The host to serve the dashboard on.", default=None
)
@click.option(
    "--port", "-p", type=click.STRING, help="The port to serve the dashboard on.", default=None
)
@click.option(
    "--debug", "-d", type=click.BOOL, help="Whether or not to run in debug mode.", default=False
)
@click.argument("storage_options", nargs=-1, type=click.UNPROCESSED)
def ui(root_dir, page_size, project_name, host, port, debug, storage_options):
    """Launch the Rubicon Dashboard."""
    if page_size != 10:
        warnings.warn("The `--page-size` option will be deprecated in a future release.")

    # convert the additional storage options into a dict
    # coming in as: ('--key1', 'one', '--key2', 'two')
    cleaned_storage_options = {
        storage_options[i][2:]: storage_options[i + 1] for i in range(0, len(storage_options), 2)
    }
    rubicon = Rubicon(persistence="filesystem", root_dir=root_dir, **cleaned_storage_options)

    if project_name is None:
        warnings.warn(
            "`--project-name` will be a required option in a future release. "
            "Visualizing the most recently logged project."
        )

        project_name = [p.name for p in rubicon.projects()][0]

    project = rubicon.get_project(name=project_name)
    dashboard = Dashboard(project.experiments())

    run_server_kwargs = dict(debug=debug, port=port, host=host)
    run_server_kwargs = {k: v for k, v in run_server_kwargs.items() if v is not None}
    dashboard.serve(run_server_kwargs=run_server_kwargs)


@cli.command(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@click.option(
    "--root-dir",
    default=os.environ.get("RUBICON_ROOT_DIR"),
    help="The absolute path to the top level folder holding the project.",
    required=False,  # due to environment default
    type=click.STRING,
)
@click.option(
    "--project-name",
    default=os.environ.get("RUBICON_PROJECT_NAME"),
    help="The name of the project to query experiments from.",
    required=False,  # due to environment default
    type=click.STRING,
)
@click.option(
    "--pp",
    default=False,
    help="Toggle pretty printing.",
    is_flag=True,
    required=False,
)
@click.argument("query", nargs=1, required=True)
def search(root_dir: str, project_name: str, pp: bool, query: str):
    """Query rubicon experiments from the command line with JSONPath syntax."""

    if project_name is None or root_dir is None:
        click.secho("No --root-dir or --project-name provided. Exiting...", fg="red")
        sys.exit(1)

    rubicon = Rubicon(persistence="filesystem", root_dir=root_dir)

    try:
        project = rubicon.get_project(name=project_name)
    except Exception as e:
        click.secho(e, fg="red")
        sys.exit(1)

    rubicon_json = RubiconJSON(projects=project)

    results = rubicon_json.search(query)
    results = [results] if not isinstance(results, list) else results
    matches = [r.value for r in results]

    click.secho(pprint.pformat(matches, indent=1) if pp else matches)


# CLI groups
if __name__ == "__main__":
    cli()



import warnings

from sklearn.pipeline import Pipeline, _name_estimators

from rubicon_ml.client.project import Project
from rubicon_ml.sklearn.estimator_logger import EstimatorLogger
from rubicon_ml.sklearn.utils import log_parameter_with_warning


class RubiconPipeline(Pipeline):
    """An extension of `sklearn.pipeline.Pipeline` that automatically
    creates a Rubicon `experiment` under the provided `project` and logs
    the pipeline's `parameters` and `metrics` to it.

    A single pipeline run will result in a single `experiment` logged with
    its corresponding `parameters` and `metrics` pulled from the pipeline's
    estimators.

    Parameters
    ----------
    project : rubicon_ml.client.Project
        The rubicon project to log to.
    steps : list
        List of (name, transform) tuples (implementing fit/transform) that
        are chained, in the order in which they are chained, with the last
        object an estimator.
    user_defined_loggers : dict, optional
        A dict mapping the estimator name to a corresponding user defined
        logger. See the example below for more details.
    experiment_kwargs : dict, optional
        Additional keyword arguments to be passed to
        `project.log_experiment()`.
    memory : str or object with the joblib.Memory interface, default=None
        Used to cache the fitted transformers of the pipeline. By default,
        no caching is performed. If a string is given, it is the path to
        the caching directory. Enabling caching triggers a clone of
        the transformers before fitting. Therefore, the transformer
        instance given to the pipeline cannot be inspected
        directly. Use the attribute ``named_steps`` or ``steps`` to
        inspect estimators within the pipeline. Caching the
        transformers is advantageous when fitting is time consuming.
        (docstring source: Scikit-Learn)
    verbose : bool, default=False
        If True, the time elapsed while fitting each step will be printed as
        it is completed. (docstring source: Scikit-Learn)
    ignore_warnings : bool, default=False
        If True, ignores warnings thrown by pipeline.

    Examples
    --------
    >>> pipeline = RubiconPipeline(
    ...     project,
    ...     [
    ...         ("vect", CountVectorizer()),
    ...         ("tfidf", TfidfTransformer()),
    ...         ("clf", SGDClassifier()),
    ...     ],
    ...     user_defined_loggers = {
    ...         "vect": FilterEstimatorLogger(
    ...             select=["input", "decode_error", "max_df"],
    ...         ),
    ...         "tfidf": FilterEstimatorLogger(ignore_all=True),
    ...         "clf": FilterEstimatorLogger(
    ...             ignore=["alpha", "penalty"],
    ...         ),
    ...     }
    ... )
    """

    def __init__(
        self,
        project,
        steps,
        user_defined_loggers={},
        experiment_kwargs={"name": "RubiconPipeline experiment"},
        memory=None,
        verbose=False,
        ignore_warnings=False,
    ):
        self.project = project
        self.user_defined_loggers = user_defined_loggers
        self.experiment_kwargs = experiment_kwargs

        self.experiment = None
        self.ignore_warnings = ignore_warnings

        super().__init__(steps, memory=memory, verbose=verbose)

    def fit(self, X, y=None, tags=None, log_fit_params=True, experiment=None, **fit_params):
        """Fit the model and automatically log the `fit_params`
        to rubicon-ml. Optionally, pass `tags` to update the experiment's
        tags.

        Parameters
        ----------
        X : iterable
            Training data. Must fulfill input requirements of first step of the pipeline.
        y : iterable, optional
            Training targets. Must fulfill label requirements for all steps of the pipeline.
        tags : list, optional
            Additional tags to add to the experiment during the fit.
        log_fit_params : bool, optional
            True to log the values passed as `fit_params` to this pipeline's experiment.
            Defaults to True.
        fit_params : dict, optional
            Additional keyword arguments to be passed to `sklearn.pipeline.Pipeline.fit()`.
        experiment: rubicon_ml.experiment.client.Experiment, optional
            The experiment to log the to. If no experiment is provided the metrics are
            logged to a new experiment with self.experiment_kwargs.

        Returns
        -------
        rubicon_ml.sklearn.Pipeline
            This `RubiconPipeline`.
        """
        with warnings.catch_warnings():
            if self.ignore_warnings:
                warnings.simplefilter("ignore")
            pipeline = super().fit(X, y, **fit_params)

            if experiment is None:
                experiment = self.project.log_experiment(**self.experiment_kwargs)
            self.experiment = experiment

            if tags is not None:
                self.experiment.add_tags(tags)

            for step_name, estimator in self.steps:
                logger = self.get_estimator_logger(step_name, estimator)
                logger.log_parameters()

            if log_fit_params:
                for name, value in fit_params.items():
                    log_parameter_with_warning(self.experiment, name, value)
        return pipeline

    def score(self, X, y=None, sample_weight=None, experiment=None):
        """Score with the final estimator and automatically
        log the results to rubicon-ml.

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step of the pipeline.
        y : iterable, optional
            Targets used for scoring. Must fulfill label requirements for all steps of the pipeline.
        sample_weight : list, optional
            If not None, this argument is passed as sample_weight keyword argument to the
            score method of the final estimator.
        experiment: rubicon_ml.experiment.client.Experiment, optional
            The experiment to log the score to. If no experiment is provided the score is logged
            to a new experiment with self.experiment_kwargs.

        Returns
        -------
        float
            Result of calling `score` on the final estimator.
        """
        with warnings.catch_warnings():
            if self.ignore_warnings:
                warnings.simplefilter("ignore")
            score = super().score(X, y, sample_weight)

            if experiment is not None:
                self.experiment = experiment
            elif self.experiment is None:
                self.experiment = self.project.log_experiment(**self.experiment_kwargs)

            logger = self.get_estimator_logger()
            logger.log_metric("score", score)

            # clear `self.experiment` to avoid duplicate metric logging errors
            self.experiment = None

        return score

    def score_samples(self, X, experiment=None):
        """Score samples with the final estimator and automatically
        log the results to rubicon-ml.

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step of the pipeline.
        experiment: rubicon_ml.experiment.client.Experiment, optional
            The experiment to log the score to. If no experiment is provided the score is logged
            to a new experiment with self.experiment_kwargs.

        Returns
        -------
        ndarray of shape (n_samples,)
            Result of calling `score_samples` on the final estimator.
        """
        with warnings.catch_warnings():
            if self.ignore_warnings:
                warnings.simplefilter("ignore")
            score_samples = super().score_samples(X)

            if experiment is not None:
                self.experiment = experiment
            elif self.experiment is None:
                self.experiment = self.project.log_experiment(**self.experiment_kwargs)

            logger = self.get_estimator_logger()
            try:
                logger.log_metric("score_samples", score_samples)
            except TypeError:
                score_samples = score_samples.tolist()

                logger.log_metric("score_samples", score_samples)

            # clear `self.experiment` to avoid duplicate metric logging errors
            self.experiment = None

            return score_samples

    def get_estimator_logger(self, step_name=None, estimator=None):
        """Get a logger for the estimator. By default, the logger will
        have the current experiment set.
        """
        logger = self.user_defined_loggers.get(step_name) or EstimatorLogger()

        logger.set_experiment(self.experiment)

        if step_name:
            logger.set_step_name(step_name)

        if estimator:
            logger.set_estimator(estimator)

        return logger

    def __getitem__(self, ind):
        """Returns a sub-pipeline with the configured rubicon-ml loggers or
        a single estimator in the pipeline.

        Indexing with an integer will return an estimator; using a slice
        returns another Pipeline instance which copies a slice of this
        Pipeline. This copy is shallow: modifying (or fitting) estimators in
        the sub-pipeline will affect the larger pipeline and vice-versa.
        However, replacing a value in `step` will not affect a copy.
        (docstring source: Scikit-Learn)
        """
        if isinstance(ind, slice):
            if ind.step not in (1, None):
                raise ValueError("Pipeline slicing only supports a step of 1")

            user_defined_loggers_slice = self._get_logger_slice(self.steps[ind])

            return self.__class__(
                self.project,
                self.steps[ind],
                user_defined_loggers_slice,
                self.experiment_kwargs,
                memory=self.memory,
                verbose=self.verbose,
            )
        try:
            _, est = self.steps[ind]
        except TypeError:
            # Not an int, try get step by name
            return self.named_steps[ind]
        return est

    def _get_logger_slice(self, steps):
        """Given a slice of estimators, returns the associated slice of loggers."""
        user_defined_loggers_slice = {}

        for name, _ in steps:
            if name in self.user_defined_loggers:
                user_defined_loggers_slice[name] = self.user_defined_loggers[name]

        return user_defined_loggers_slice


def make_pipeline(
    project,
    *steps,
    experiment_kwargs={"name": "RubiconPipeline experiment"},
    memory=None,
    verbose=False,
):
    """Wrapper around RubicionPipeline(). Does not require naming for estimators.
    Their names are set to the lowercase strings of their types.

    Parameters
    ----------
    project : rubicon_ml.client.Project
        The rubicon project to log to.
    steps : list
        List of  estimator objects or (estimator, logger) tuples (implementing
        fit/transform) that are chained, in the order in which they are chained,
        with the last object an estimator. (docstring source: Scikit-Learn)
    experiment_kwargs : dict, optional
        Additional keyword arguments to be passed to
        `project.log_experiment()`.
    memory : str or object with the joblib.Memory interface, default=None
        Used to cache the fitted transformers of the pipeline. By default,
        no caching is performed. If a string is given, it is the path to
        the caching directory. Enabling caching triggers a clone of
        the transformers before fitting. Therefore, the transformer
        instance given to the pipeline cannot be inspected
        directly. Use the attribute ``named_steps`` or ``steps`` to
        inspect estimators within the pipeline. Caching the
        transformers is advantageous when fitting is time consuming.
        (docstring source: Scikit-Learn)
    verbose : bool, default=False
        If True, the time elapsed while fitting each step will be printed as it
        is completed. (docstring source: Scikit-Learn)

    Returns
    -------
    rubicon_ml.sklearn.Pipeline
        A `RubiconPipeline` with project `project` and steps `steps`.
    """
    steps, loggers = _split_steps_loggers(steps)

    steps = _name_estimators(steps)
    user_defined_loggers = _name_loggers(steps, loggers)

    if not isinstance(project, Project):
        raise ValueError(
            "project" + str(project) + " must be of type rubicon_ml.client.project.Project"
        )

    return RubiconPipeline(project, steps, user_defined_loggers, experiment_kwargs, memory, verbose)


def _split_steps_loggers(steps):
    """Splits the loggers and returns the estimators in the format
    scikit-learn expects them.

    Parameters
    ----------
    steps: list of sklearn.Estimator or tuples of (sklearn.Estimator,
    rubicon_ml.sklearn.EstimatorLogger).
        The steps and estimator loggers to split.

    Returns
    -------
    list of sklearn.Estimator and list of rubicon_ml.sklearn.EstimatorLogger
        The ordered lists of estimators and rubicon-ml loggers.
    """
    ret_loggers = []
    ret_steps = []

    for step in steps:
        if isinstance(step, tuple):
            ret_loggers.append(step[1])
            ret_steps.append(step[0])
        else:
            ret_loggers.append(None)
            ret_steps.append(step)

    return ret_steps, ret_loggers


def _name_loggers(steps, loggers):
    """Names loggers in the format rubicon-ml expects them.

    Parameters
    ----------
    steps: list of tuples of (str, sklearn.Estimator)
        The named estimator steps.
    loggers: list of rubicon_ml.sklearn.EstimatorLogger
        The rubicon-ml loggers.

    Returns
    -------
    dict of str, rubicon_ml.sklearn.EstimatorLogger
        The named rubicon-ml loggers.
    """
    named_loggers = {}

    for i in range(len(steps)):
        if loggers[i] is not None:
            named_loggers[str(steps[i][0])] = loggers[i]

    return named_loggers



"""Methods and a mixin to enable schema logging.

The functions available in the ``schema`` submodule are applied to
``rubicon_ml.Project`` via the ``SchemaMixin`` class. They can be
called directly as a method of an existing project.
"""

import os
from contextlib import contextmanager
from typing import Any, Dict, Optional

from rubicon_ml.client.experiment import Experiment
from rubicon_ml.exceptions import RubiconException
from rubicon_ml.schema import registry


def _get_value(obj, entity_schema):
    optional = entity_schema.get("optional", False)
    value = None

    if "value_attr" in entity_schema:
        value = _safe_getattr(obj, entity_schema["value_attr"], optional)
    if "value_env" in entity_schema:
        value = _safe_environ(entity_schema["value_env"], optional)
    if "value_func" in entity_schema:
        value = _safe_call_func(obj, entity_schema["value_func"], optional)

    return value


def _get_df(obj, entity_schema):
    optional = entity_schema.get("optional", False)
    df_value = None

    if "df_attr" in entity_schema:
        df_value = _safe_getattr(obj, entity_schema["df_attr"], optional)
    if "df_func" in entity_schema:
        df_value = _safe_call_func(obj, entity_schema["df_func"], optional)

    return df_value


def _get_data_object(obj, entity_schema):
    optional = entity_schema.get("optional", False)
    data_object = None

    if "data_object_func" in entity_schema:
        data_object = _safe_call_func(obj, entity_schema["data_object_func"], optional)
    elif "data_object_attr" in entity_schema:
        data_object = _safe_getattr(obj, entity_schema["data_object_attr"], optional)

    return data_object


def _safe_getattr(obj, attr, optional, default=None):
    try:
        value = getattr(obj, attr)
    except (TypeError, AttributeError) as err:
        if optional or (attr is None and isinstance(err, TypeError)):
            return default

        raise err

    return value


def _safe_environ(environ_var, optional, default=None):
    try:
        value = os.environ[environ_var]
    except KeyError as err:
        if optional:
            return default

        raise RubiconException(f"Environment variable '{environ_var}' not set.") from err

    return value


def _safe_call_func(obj, func, optional, default=None):
    method = _safe_getattr(obj, func, optional, default)
    value = None

    if method is not None:
        try:
            value = method()
        except Exception as err:
            if optional:
                return default

            raise err

    return value


@contextmanager
def _set_temporary_schema(project, schema_name):
    original_schema = project.schema_

    if schema_name == "infer":
        delattr(project, "schema_")
    else:
        project.set_schema(registry.get_schema(schema_name))

    yield

    project.set_schema(original_schema)


class SchemaMixin:
    """Adds schema logging support to a client object."""

    def log_with_schema(
        self,
        obj: Any,
        experiment: Experiment = None,
        experiment_kwargs: Optional[Dict[str, Any]] = None,
    ) -> Any:
        """Log an experiment leveraging ``self.schema_``."""

        if not hasattr(self, "schema_"):
            try:
                schema_name = registry.get_schema_name(obj)
                self.schema_ = registry.get_schema(schema_name)
            except ValueError as err:
                raise ValueError(
                    f"No schema set and no schema could be inferred from object {obj}. "
                    f"Set a schema with `Project.set_schema(schema)`."
                ) from err

        if experiment_kwargs is None:
            experiment_kwargs = {}

        if experiment is None:
            experiment = self.log_experiment(**experiment_kwargs)

        base_schema_name = self.schema_.get("extends")
        if base_schema_name is not None:
            with _set_temporary_schema(self, base_schema_name):
                self.log_with_schema(obj, experiment=experiment)

        for feature in self.schema_.get("features", []):
            is_optional = feature.get("optional", False)

            if "names_attr" in feature:
                feature_names = _safe_getattr(obj, feature["names_attr"], is_optional)

                if feature_names is not None:
                    feature_importances = _safe_getattr(
                        obj,
                        feature.get("importances_attr"),
                        is_optional,
                        default=[None] * len(feature["names_attr"]),
                    )

                    for name, importance in zip(feature_names, feature_importances):
                        experiment.log_feature(name=name, importance=importance)

            elif "name_attr" in feature:
                feature_name = _safe_getattr(obj, feature["name_attr"], is_optional)

                if feature_name is not None:
                    feature_importance = _safe_getattr(
                        obj, feature.get("importance_attr"), is_optional
                    )

                    experiment.log_feature(name=feature_name, importance=feature_importance)

        for parameter in self.schema_.get("parameters", []):
            experiment.log_parameter(
                name=parameter["name"],
                value=_get_value(obj, parameter),
            )

        for metric in self.schema_.get("metrics", []):
            experiment.log_metric(
                name=metric["name"],
                value=_get_value(obj, metric),
            )
        for artifact in self.schema_.get("artifacts", []):
            if isinstance(artifact, str):
                if artifact == "self":
                    experiment.log_artifact(name=obj.__class__.__name__, data_object=obj)
            elif isinstance(artifact, dict):
                if "self" in artifact:
                    logging_func_name = artifact["self"]
                    logging_func = getattr(experiment, logging_func_name)

                    # Get remaining artifact logging function parameters and run with func
                    logging_func(
                        obj, **dict((k, v) for k, v in artifact.items() if k != "self")
                    )  # key-values in rest of dictionary are passed as arguments
                else:
                    data_object = _get_data_object(obj, artifact)

                    if data_object is not None:
                        experiment.log_artifact(name=artifact["name"], data_object=data_object)

        for dataframe in self.schema_.get("dataframes", []):
            df_value = _get_df(obj, dataframe)

            if df_value is not None:
                experiment.log_dataframe(df=df_value, name=dataframe["name"])

        for schema in self.schema_.get("schema", []):
            object_to_log = _safe_getattr(obj, schema["attr"], schema.get("optional", False))

            if object_to_log is not None:
                with _set_temporary_schema(self, schema["name"]):
                    self.log_with_schema(object_to_log, experiment=experiment)

        has_children = False

        for children in self.schema_.get("children", []):
            children_objects = _safe_getattr(
                obj, children["attr"], children.get("optional", False), default=[]
            )

            for child in children_objects:
                has_children = True

                child_experiment = self.log_experiment(**experiment_kwargs)
                child_experiment.add_tags(tags=["child", f"parent_id:{experiment.id}"])

                with _set_temporary_schema(self, children["name"]):
                    self.log_with_schema(child, experiment=child_experiment)

        if has_children:
            experiment.add_tags(tags=["parent"])

        return experiment

    def set_schema(self, schema: Dict[str, Any]) -> None:
        """Set the schema for this client object."""

        self.schema_ = schema



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Optional, Union

from rubicon_ml.client import Base, CommentMixin, TagMixin

if TYPE_CHECKING:
    from rubicon_ml.client import Experiment
    from rubicon_ml.domain import Parameter as ParameterDomain


class Parameter(Base, TagMixin, CommentMixin):
    """A client parameter.

    A `parameter` is an input to an `experiment` (model run)
    that depends on the type of model being used. It affects
    the model's predictions.

    For example, if you were using a random forest classifier,
    'n_estimators' (the number of trees in the forest) could
    be a parameter.

    A `parameter` is logged to an `experiment`.

    Parameters
    ----------
    domain : rubicon.domain.Parameter
        The parameter domain model.
    parent : rubicon.client.Experiment
        The experiment that the parameter is logged to.
    """

    def __init__(self, domain: ParameterDomain, parent: Experiment):
        super().__init__(domain, parent._config)
        self._parent = parent

        self._domain: ParameterDomain

    @property
    def id(self) -> str:
        """Get the parameter's id."""
        return self._domain.id

    @property
    def name(self) -> Optional[str]:
        """Get the parameter's name."""
        return self._domain.name

    @property
    def value(self) -> Optional[Union[object, float]]:
        """Get the parameter's value."""
        return self._domain.value

    @property
    def description(self) -> Optional[str]:
        """Get the parameter's description."""
        return self._domain.description

    @property
    def created_at(self) -> datetime:
        """Get the time the parameter was created."""
        return self._domain.created_at

    @property
    def parent(self) -> Experiment:
        """Get the parameter's parent client object."""
        return self._parent



import copy
import json

import numpy as np
import plotly.figure_factory as ff
from dash import callback_context, dcc, html
from dash.dependencies import ALL, Input, Output

from rubicon_ml.viz.base import VizBase
from rubicon_ml.viz.common import dropdown_header
from rubicon_ml.viz.common.colors import light_blue, plot_background_blue


class MetricListsComparison(VizBase):
    """Visualize lists of metrics logged to the experiments `experiments` as
    an annotated heatmap.

    More info on annotated heatmaps can be found here:
    https://plotly.com/python/annotated-heatmap/

    Parameters
    ----------
    column_names : list of str
        Titles to use for each column in the heatmap. Defaults to None.
    experiments : list of rubicon_ml.client.experiment.Experiment, optional
        The experiments to visualize. Defaults to None. Can be set as
        attribute after instantiation.
    selected_metric : str
        The name of the metric to display at launch. Defaults to None, which
        selects the metric loaded first.
    """

    def __init__(
        self,
        column_names=None,
        experiments=None,
        selected_metric=None,
    ):
        super().__init__(dash_title="compare metric lists")

        if column_names is None:
            self.column_names = []
        else:
            self.column_names = column_names
        self.experiments = experiments
        self.selected_metric = selected_metric

    @property
    def layout(self):
        """Defines the layout for the metric lists comparison."""
        return html.Div(
            [
                dropdown_header(
                    list(self.metric_names),
                    self.selected_metric,
                    "comparing metric ",
                    f" over {len(self.experiments)} experiments",
                    "metric",
                ),
                dcc.Loading(
                    html.Div(
                        dcc.Graph(
                            id="metric-heatmap",
                        ),
                        id="metric-heatmap-container",
                    ),
                    color=light_blue,
                ),
            ],
            id="metric-heatmap-layout-container",
        )

    def load_experiment_data(self):
        """Load the experiment data required for the metric lists comparison.

        Extracts all metric metadata from each experiment in `self.experiment`
        if the metric's value is a list.
        """
        self.experiment_records = {}
        self.metric_names = set()

        for experiment in self.experiments:
            for metric in experiment.metrics():
                if isinstance(metric.value, list):
                    self.metric_names.add(metric.name)

                    experiment_record = self.experiment_records.get(experiment.id, {})
                    experiment_record[metric.name] = metric.value
                    self.experiment_records[experiment.id] = experiment_record

                    if self.selected_metric is None:
                        self.selected_metric = metric.name

        if self.selected_metric not in self.metric_names:
            raise ValueError(
                f"no metric named `selected_metric` '{self.selected_metric}'"
                " logged to any experiment in `experiments`."
            )

    def register_callbacks(self, link_experiment_table=False):
        outputs = [
            Output("metric-heatmap", "figure"),
            Output("metric-heatmap", "style"),
            Output("metric-header-right-text", "children"),
            Output("metric-dropdown", "label"),
        ]
        inputs = [Input({"type": "metric-dropdown-button", "index": ALL}, "n_clicks")]
        states = []

        if link_experiment_table:
            inputs.append(
                Input("experiment-table", "derived_virtual_selected_row_ids"),
            )

        @self.app.callback(outputs, inputs, states)
        def update_metric_heatmap(*args):
            """Render the heatmap based on the currently selected metric.

            Returns the Plotly annotated heatmap generated by the values of the
            experiments' selected metric, the heatmap's style, the header text
            with the metric's name, and the selected metric's name.
            """
            if link_experiment_table:
                selected_row_ids = args[-1]
                selected_row_ids = selected_row_ids if selected_row_ids else []
            else:
                selected_row_ids = self.experiment_records.keys()

            property_id = callback_context.triggered[0].get("prop_id")
            property_value = property_id[: property_id.index(".")]

            if not property_value or property_value == "experiment-table":
                selected_metric = self.selected_metric
            else:
                selected_metric = json.loads(property_value).get("index")

                self.selected_metric = selected_metric

            heatmap_data = []
            experiment_ids = []

            for experiment_id, experiment_record in self.experiment_records.items():
                if experiment_id in selected_row_ids:
                    metric_value = experiment_record.get(selected_metric)

                    if metric_value is not None:
                        heatmap_data.append(metric_value)
                        experiment_ids.append(experiment_id[:7])

            header_right_text = (
                f"over {len(experiment_ids)} experiment{'s' if len(experiment_ids) != 1 else ''}"
            )

            if len(heatmap_data) == 0:
                return [], {"display": "none"}, header_right_text, selected_metric

            data_array = np.array(heatmap_data)
            numerator = data_array - data_array.min(axis=0)
            denominator = data_array.max(axis=0) - data_array.min(axis=0)
            denominator[denominator == 0] = 1
            scaled_heatmap_data = numerator / denominator

            annotations = copy.deepcopy(heatmap_data)
            for i, row in enumerate(annotations):
                for j, label in enumerate(row):
                    if isinstance(label, float):
                        annotations[i][j] = round(label, 6)

            heatmap_margin = 30
            heatmap_min_width = 72
            heatmap_cell_width = 6
            heatmap_cell_height = heatmap_cell_width / 2
            heatmap_cell_buffer = 12

            heatmap = ff.create_annotated_heatmap(
                scaled_heatmap_data,
                annotation_text=annotations,
                colorscale="blues",
                hoverinfo="text",
                text=heatmap_data,
                x=self.column_names if len(self.column_names) == len(heatmap_data[0]) else None,
                y=experiment_ids,
            )
            heatmap.update_layout(
                margin_b=heatmap_margin,
                margin_t=heatmap_margin,
                modebar_orientation="v",
                plot_bgcolor=plot_background_blue,
            )
            heatmap.update_xaxes(gridcolor="white")
            heatmap.update_yaxes(gridcolor="white")

            heatmap_height = heatmap_cell_buffer + (len(heatmap_data) * heatmap_cell_height)
            heatmap_width = (
                heatmap_cell_buffer + (len(heatmap_data[0]) * heatmap_cell_width)
                if len(heatmap_data[0]) > 8
                else heatmap_min_width
            )
            heatmap_style = {"height": f"{heatmap_height}rem", "width": f"{heatmap_width}rem"}

            return heatmap, heatmap_style, header_right_text, selected_metric



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional, Union

from rubicon_ml.exceptions import RubiconException

if TYPE_CHECKING:
    from rubicon_ml.client import Config
    from rubicon_ml.domain import DOMAIN_TYPES
    from rubicon_ml.repository import BaseRepository


class Base:
    """The base object for all top-level client objects.

    Parameters
    ----------
    domain : one of rubicon.domain.*
        The top-level object's domain instance.
    config : rubicon.client.Config, optional
        The config, which injects the repository to use.
    """

    def __init__(self, domain: DOMAIN_TYPES, config: Optional[Union[Config, List[Config]]] = None):
        self._config = config
        self._domain = domain

    def __str__(self) -> str:
        return self._domain.__str__()

    def is_auto_git_enabled(self) -> bool:
        """Is git enabled for any of the configs."""
        if isinstance(self._config, list):
            return any(_config.is_auto_git_enabled for _config in self._config)

        if self._config is None:
            return False

        return self._config.is_auto_git_enabled

    def _raise_rubicon_exception(self, exception: Exception):
        if self.repositories is None or len(self.repositories) > 1:
            raise RubiconException("all configured storage backends failed") from exception
        else:
            raise exception

    @property
    def repository(self) -> Optional[BaseRepository]:
        """Get the repository."""
        if self._config is None:
            return None

        if isinstance(self._config, list):
            if len(self._config) > 1:
                raise ValueError("More than one repository available. Use `.repositories` instead.")
            return self._config[0].repository

        return self._config.repository

    @property
    def repositories(self) -> Optional[List[BaseRepository]]:
        """Get all repositories."""
        if self._config is None:
            return None

        if isinstance(self._config, list):
            return [_config.repository for _config in self._config]

        return [self._config.repository]



import string


def slugify(value):
    # Remove punctuation
    value = value.translate(str.maketrans("", "", string.punctuation))
    # Remove any excess whitespace within the string
    value = " ".join(value.split()).strip()
    # Lowercase and replace spaces with our separator
    return value.lower().replace(" ", "-")



from rubicon_ml import __version__
from rubicon_ml.intake_rubicon.base import VizDataSourceMixin


class ExperimentsTableDataSource(VizDataSourceMixin):
    """An Intake data source for reading `rubicon` Experiment Table visualizations."""

    version = __version__

    container = "python"
    name = "rubicon_ml_experiments_table"

    def __init__(self, metadata=None, **catalog_data):
        self._catalog_data = catalog_data or {}

        super().__init__(metadata=metadata)

    def _get_schema(self):
        """Creates an Experiments Table visualization and sets it as the visualization object attribute"""
        from rubicon_ml.viz import ExperimentsTable

        self._visualization_object = ExperimentsTable(**self._catalog_data)

        return super()._get_schema()


class MetricCorrelationPlotDataSource(VizDataSourceMixin):
    """An Intake data source for reading `rubicon` Metric Correlation Plot visualizations."""

    version = __version__

    container = "python"
    name = "rubicon_ml_metric_correlation_plot"

    def __init__(self, metadata=None, **catalog_data):
        self._catalog_data = catalog_data or {}

        super().__init__(metadata=metadata)

    def _get_schema(self):
        """Creates a Metric Correlation Plot visualization and sets it as the visualization object attribute"""
        from rubicon_ml.viz import MetricCorrelationPlot

        self._visualization_object = MetricCorrelationPlot(**self._catalog_data)

        return super()._get_schema()


class DataframePlotDataSource(VizDataSourceMixin):
    """An Intake data source for reading `rubicon` Dataframe Plot visualizations."""

    version = __version__

    container = "python"
    name = "rubicon_ml_dataframe_plot"

    def __init__(self, metadata=None, **catalog_data):
        self._catalog_data = catalog_data or {}

        super().__init__(metadata=metadata)

    def _get_schema(self):
        """Creates a Dataframe Plot visualization and sets it as the visualization object attribute"""
        from rubicon_ml.viz import DataframePlot

        self._visualization_object = DataframePlot(**self._catalog_data)

        return super()._get_schema()


class MetricListComparisonDataSource(VizDataSourceMixin):
    """An Intake data source for reading `rubicon` Metric List Comparison visualizations."""

    version = __version__

    container = "python"
    name = "rubicon_ml_metric_list"

    def __init__(self, metadata=None, **catalog_data):
        self._catalog_data = catalog_data or {}

        super().__init__(metadata=metadata)

    def _get_schema(self):
        """Creates a Metric List Comparison visualization and sets it as the visualization object attribute"""
        from rubicon_ml.viz import MetricListsComparison

        self._visualization_object = MetricListsComparison(**self._catalog_data)

        return super()._get_schema()



import os
import shutil
import tempfile
import warnings
from datetime import datetime
from json import JSONDecodeError
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional, Type, Union
from zipfile import ZipFile

import fsspec
import pandas as pd

from rubicon_ml import domain
from rubicon_ml.domain.utils.uuid import uuid4
from rubicon_ml.exceptions import RubiconException
from rubicon_ml.repository.utils import json, slugify

if TYPE_CHECKING:
    import dask.dataframe as dd
    import polars as pl


class BaseRepository:
    """The base repository defines all the shared interactions
    between the various Rubicon persistence options.

    `BaseRepositoy` itself should never be used directly. Use
    one of the repositories that extends this class to persist
    Rubicon data:

        * `rubicon.repository.MemoryRepository`
        * `rubicon.repository.LocalRepository`
        * `rubicon.repository.S3Repository`

    Parameters
    ----------
    root_dir : str
        Absolute path to the root directory to persist Rubicon
        data to.
    storage_options : dict, optional
        Additional keyword arguments that are passed directly to
        the underlying filesystem class.
    """

    def __init__(self, root_dir: str, **storage_options):
        self._df_storage_options = {}  # should only be non-empty for S3 logging

        self.filesystem = fsspec.filesystem(self.PROTOCOL, **storage_options)
        self.root_dir = root_dir.rstrip("/")

    # --- Filesystem Helpers ---

    def _cat(self, path: str):
        """Returns the contents of the file at `path`."""
        return self.filesystem.cat(path)

    def _cat_paths(self, metadata_paths: List[str]) -> Dict[str, Any]:
        """Cat `metadata_paths` to get the list of files to include.
        Ignore FileNotFoundErrors to avoid misc file errors, like hidden
        dotfiles.
        """
        if not isinstance(metadata_paths, list):
            metadata_paths = [metadata_paths]
        if not metadata_paths:
            return {}

        files = {}

        for path, metadata in self.filesystem.cat(metadata_paths, on_error="return").items():
            if isinstance(metadata, FileNotFoundError):
                warning = f"{path} not found. Was this file unintentionally created?"
                warnings.warn(warning)
            else:
                files[path] = metadata

        return files

    def _exists(self, path: str) -> bool:
        """Returns True if a file exists at `path`, False otherwise."""
        return self.filesystem.exists(path)

    def _glob(self, globstring: str):
        """Returns the names of the files matching `globstring`."""
        return self.filesystem.glob(globstring, detail=True)

    def _ls_directories_only(self, path: str) -> List[str]:
        """Returns the names of all the directories at path `path`."""
        directories = [
            os.path.join(p.get("name"), "metadata.json")
            for p in self.filesystem.ls(path, detail=True)
            if p.get("type", p.get("StorageClass")).lower() == "directory"
        ]

        return directories

    def _ls(self, path: str):
        return self.filesystem.ls(path)

    def _mkdir(self, dirpath: str):
        """Creates a directory `dirpath` with parents."""
        return self.filesystem.mkdirs(dirpath, exist_ok=True)

    def _modified(self, path: str):
        return self.filesystem.modified(path)

    def _persist_bytes(self, bytes_data, path):
        """Write bytes to the filesystem.

        To be implemented by extensions of the base filesystem.
        """
        raise NotImplementedError()

    def _persist_domain(self, domain, path):
        """Write a domain object to the filesystem.

        To be implemented by extensions of the base filesystem.
        """
        raise NotImplementedError()

    def _read_bytes(self, path, err_msg=None):
        """Read bytes from the file at `path`."""
        try:
            open_file = self.filesystem.open(path, "rb")
        except FileNotFoundError:
            raise RubiconException(err_msg)

        return open_file.read()

    def _read_domain(self, path, err_msg=None):
        """Read a domain object from the file at `path`."""
        try:
            open_file = self.filesystem.open(path)
        except FileNotFoundError:
            raise RubiconException(err_msg)

        return json.load(open_file)

    def _rm(self, path):
        """Recursively remove all files at `path`."""
        return self.filesystem.rm(path, recursive=True)

    def _load_metadata_files(
        self, metadata_root: str, domain_type: Type[domain.DomainsVar]
    ) -> List[domain.DomainsVar]:
        """Load metadata files from the given root directory and return a list of domain objects."""
        # find all directories, prepare a list of those plus `metadata.yaml`
        try:
            metadata_paths = self._ls_directories_only(metadata_root)
        except FileNotFoundError:
            return []

        loaded_domains = []
        # cat_paths will check for FileNotFoundErrors and skip any missing files
        # it loads the contents of the found files
        for path, metadata in self._cat_paths(metadata_paths).items():
            try:
                metadata_contents = json.loads(metadata)
            except JSONDecodeError:
                warnings.warn(f"Failed to load metadata for {domain_type.__name__} at {path}")
                continue

            try:
                loaded_domain = domain_type(**metadata_contents)
            except TypeError:
                warnings.warn(f"Failed to load {domain_type.__name__} from metadata at {path}")
                continue

            loaded_domains.append(loaded_domain)

        if loaded_domains:
            loaded_domains.sort(key=lambda d: d.created_at)

        return loaded_domains

    # -------- Projects --------

    def _get_project_metadata_path(self, project_name: str):
        """Returns the path of the project with name `project_name`'s
        metadata.
        """
        return f"{self.root_dir}/{slugify(project_name)}/metadata.json"

    def create_project(self, project: domain.Project):
        """Persist a project to the configured filesystem.

        Parameters
        ----------
        project : rubicon.domain.Project
            The project to persist.
        """
        project_metadata_path = self._get_project_metadata_path(project.name)

        if self._exists(project_metadata_path):
            raise RubiconException(f"A project with name '{project.name}' already exists.")

        self._persist_domain(project, project_metadata_path)

    def get_project(self, project_name: str) -> domain.Project:
        """Retrieve a project from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project to retrieve.

        Returns
        -------
        rubicon.domain.Project
            The project with name `project_name`.
        """
        project_metadata_path = self._get_project_metadata_path(project_name)
        project = self._read_domain(
            project_metadata_path,
            f"No project with name '{project_name}' found.",
        )

        return domain.Project(**project)

    def get_projects(self) -> List[domain.Project]:
        """Get the list of projects from the filesystem.

        Returns
        -------
        list of rubicon.domain.Project
            The list of projects from the filesystem.
        """
        return self._load_metadata_files(self.root_dir, domain.Project)

    # ------ Experiments -------

    def _get_experiment_metadata_root(self, project_name: str):
        """Returns the experiments directory of the project with
        name `project_name`.
        """
        return f"{self.root_dir}/{slugify(project_name)}/experiments"

    def _get_experiment_metadata_path(self, project_name, experiment_id):
        """Returns the path of the experiment with ID
        `experiment_id`'s metadata.
        """
        experiment_metadata_root = self._get_experiment_metadata_root(project_name)

        return f"{experiment_metadata_root}/{experiment_id}/metadata.json"

    def create_experiment(self, experiment: domain.Experiment):
        """Persist an experiment to the configured filesystem.

        Parameters
        ----------
        experiment : rubicon.domain.Experiment
            The experiment to persist.
        """
        experiment_metadata_path = self._get_experiment_metadata_path(
            experiment.project_name, experiment.id
        )

        self._persist_domain(experiment, experiment_metadata_path)

    def get_experiment(self, project_name: str, experiment_id: str) -> domain.Experiment:
        """Retrieve an experiment from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the experiment with ID
            `experiment_id` is logged to.
        experiment_id : str
            The ID of the experiment to retrieve.

        Returns
        -------
        rubicon.domain.Experiment
            The experiment with ID `experiment_id`.
        """
        experiment_metadata_path = self._get_experiment_metadata_path(project_name, experiment_id)
        experiment = self._read_domain(
            experiment_metadata_path,
            f"No experiment with id `{experiment_id}` found.",
        )

        return domain.Experiment(**experiment)

    def get_experiments(self, project_name: str) -> List[domain.Experiment]:
        """Retrieve all experiments from the configured filesystem
        that belong to the project with name `project_name`.

        Parameters
        ----------
        project_name : str
            The name of the project to retrieve all experiments
            from.

        Returns
        -------
        list of rubicon.domain.Experiment
            The experiments logged to the project with name
            `project_name`.
        """
        experiment_metadata_root = self._get_experiment_metadata_root(project_name)

        return self._load_metadata_files(experiment_metadata_root, domain.Experiment)

    # ------- Archiving --------

    def _archive(
        self,
        project_name,
        experiments: Optional[List] = None,
        remote_rubicon_root: Optional[str] = None,
    ):
        """Archive the experiments logged to this project.

        Parameters
        ----------
        project_name : str
            Name of the calling project (project to create archive for)
        experiments : list of Experiments, optional
            The rubicon.client.Experiment objects to archive. If None all logged experiments are archived.
        remote_rubicon_root : str or pathlike object, optional
            The remote root of the repository to archive to

        Returns
        -------
        filepath of newly created archive
        """
        remote_s3 = True if remote_rubicon_root and remote_rubicon_root.startswith("s3") else False
        root_dir = remote_rubicon_root if remote_rubicon_root is not None else self.root_dir
        archive_dir = os.path.join(root_dir, slugify(project_name), "archives")
        ts = datetime.timestamp(datetime.now())
        archive_path = os.path.join(archive_dir, "archive-" + str(ts))
        zip_archive_filename = str(archive_path + ".zip")
        experiments_path = self._get_experiment_metadata_root(project_name)

        if not remote_s3:
            if not self._exists(archive_dir):
                self._mkdir(archive_dir)

        file_name = None
        with tempfile.NamedTemporaryFile() as tf:
            if experiments is not None:
                with ZipFile(tf, "x") as archive:
                    experiment_paths = []
                    for experiment in experiments:
                        experiment_paths.append(os.path.join(experiments_path, experiment.id))
                    for file_path in experiment_paths:
                        archive.write(file_path, os.path.basename(file_path))
                file_name = archive.filename

            else:
                file_name = shutil.make_archive(tf.name, "zip", experiments_path)

            with fsspec.open(zip_archive_filename, "wb") as fp:
                with open(file_name, "rb") as tf:
                    fp.write(tf.read())

        return zip_archive_filename

    def _experiments_from_archive(
        self,
        project_name,
        remote_rubicon_root: str,
        latest_only: Optional[bool] = False,
    ):
        """Retrieve archived experiments into this project's experiments folder.

        Parameters
        ----------
        project_name : str
            Name of the calling project (project to read experiments into)
        remote_rubicon_root : str or pathlike object
            The remote Rubicon object with the repository containing archived experiments to read in
        latest_only : bool, optional
            Indicates whether or not experiments should only be read from the latest archive
        """
        root_dir = self.root_dir
        shutil.copy(
            os.path.join(remote_rubicon_root, slugify(project_name), "metadata.json"),
            os.path.join(root_dir, slugify(project_name)),
        )
        archive_dir = os.path.join(remote_rubicon_root, slugify(project_name), "archives")
        if not self._exists(archive_dir):
            raise ValueError("`remote_rubicon_root` has no archives")

        dest_experiments_dir = self._get_experiment_metadata_root(project_name)
        if not self._exists(dest_experiments_dir):
            self._mkdir(dest_experiments_dir)

        og_num_experiments = len(self._ls(dest_experiments_dir))

        if not latest_only:
            for zip_archive_name in self._ls(archive_dir):
                zip_archive_filepath = os.path.join(archive_dir, zip_archive_name)
                with ZipFile(zip_archive_filepath, "r") as curr_archive:
                    curr_archive.extractall(dest_experiments_dir)
        else:
            latest_zip_archive_filepath = None
            latest_time = None
            for zip_archive in self._ls(archive_dir):
                zip_archive_filepath = os.path.join(archive_dir, zip_archive)
                mod_time = self._modified(zip_archive_filepath)
                if latest_time is None:
                    latest_time = mod_time
                    latest_zip_archive_filepath = zip_archive_filepath
                elif mod_time > latest_time:
                    latest_zip_archive_filepath = zip_archive_filepath
                    latest_time = mod_time
            with ZipFile(latest_zip_archive_filepath, "r") as zip_archive:
                zip_archive.extractall(dest_experiments_dir)

        if len(self._ls(dest_experiments_dir)) > og_num_experiments:
            print("experiments read from archive")
        else:
            print("experiments not read from archive")

    # ------- Artifacts --------

    def _get_artifact_metadata_root(self, project_name, experiment_id=None):
        """Returns the artifacts directory of the project with name
        `project_name` or experiment with ID `experiment_id`.
        """
        if experiment_id is not None:
            experiment_metadata_root = self._get_experiment_metadata_root(project_name)

            return f"{experiment_metadata_root}/{experiment_id}/artifacts"
        else:
            return f"{self.root_dir}/{slugify(project_name)}/artifacts"

    def _get_artifact_metadata_path(self, project_name, experiment_id, artifact_id):
        """Returns the path of the artifact with ID `artifact_id`'s
        metadata.
        """
        artifact_metadata_root = self._get_artifact_metadata_root(project_name, experiment_id)

        return f"{artifact_metadata_root}/{artifact_id}/metadata.json"

    def _get_artifact_data_path(self, project_name, experiment_id, artifact_id):
        """Returns the path of the artifact with ID `artifact_id`'s
        raw data.
        """
        artifact_metadata_root = self._get_artifact_metadata_root(project_name, experiment_id)

        return f"{artifact_metadata_root}/{artifact_id}/data"

    def create_artifact(self, artifact, data, project_name, experiment_id=None):
        """Persist an artifact to the configured filesystem.

        Parameters
        ----------
        artifact : rubicon.domain.Artifact
            The artifact to persist.
        data : bytes
            The raw data to persist as an artifact.
        project_name : str
            The name of the project this artifact belongs to.
        experiment_id : str, optional
            The ID of the experiment this artifact belongs to.
            Artifacts do not need to belong to an experiment.
        """
        artifact_metadata_path = self._get_artifact_metadata_path(
            project_name, experiment_id, artifact.id
        )
        artifact_data_path = self._get_artifact_data_path(project_name, experiment_id, artifact.id)

        self._persist_bytes(data, artifact_data_path)
        self._persist_domain(artifact, artifact_metadata_path)

    def get_artifact_metadata(self, project_name, artifact_id, experiment_id=None):
        """Retrieve an artifact's metadata from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the artifact with ID
            `artifact_id` is logged to.
        artifact_id : str
            The ID of the artifact to retrieve.
        experiment_id : str, optional
            The ID of the experiment the artifact with ID
            `artifact_id` is logged to. Artifacts do not
            need to belong to an experiment.

        Returns
        -------
        rubicon.domain.Artifact
            The artifact with ID `artifact_id`.
        """
        artifact_metadata_path = self._get_artifact_metadata_path(
            project_name, experiment_id, artifact_id
        )
        artifact = self._read_domain(
            artifact_metadata_path,
            f"No artifact with id `{artifact_id}` found.",
        )

        return domain.Artifact(**artifact)

    def get_artifacts_metadata(self, project_name, experiment_id=None):
        """Retrieve all artifacts' metadata from the configured
        filesystem that belong to the specified object.

        Parameters
        ----------
        project_name : str
            The name of the project to retrieve all artifacts
            from.
        experiment_id : str, optional
            The ID of the experiment to retrieve all artifacts
            from. Artifacts do not need to belong to an
            experiment.

        Returns
        -------
        list of rubicon.domain.Artifact
            The artifacts logged to the specified object.
        """
        artifact_metadata_root = self._get_artifact_metadata_root(project_name, experiment_id)

        return self._load_metadata_files(artifact_metadata_root, domain.Artifact)

    def get_artifact_data(self, project_name, artifact_id, experiment_id=None):
        """Retrieve an artifact's raw data.

        Parameters
        ----------
        project_name : str
            The name of the project the artifact with ID
            `artifact_id` is logged to.
        artifact_id : str
            The ID of the artifact to retrieve data from.
        experiment_id : str, optional
            The ID of the experiment the artifact with ID
            `artifact_id` is logged to. Artifacts do not
            need to belong to an experiment.

        Returns
        -------
        bytes
            The artifact with ID `artifact_id`'s raw data.
        """
        artifact_data_path = self._get_artifact_data_path(project_name, experiment_id, artifact_id)
        artifact_data = self._read_bytes(
            artifact_data_path,
            f"No data for artifact with id `{artifact_id}` found.",
        )

        return artifact_data

    def delete_artifact(self, project_name, artifact_id, experiment_id=None):
        """Delete an artifact from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the artifact with ID
            `artifact_id` is logged to.
        artifact_id : str
            The ID of the artifact to delete.
        experiment_id : str, optional
            The ID of the experiment the artifact with ID
            `artifact_id` is logged to. Artifacts do not
            need to belong to an experiment.
        """
        artifact_metadata_root = self._get_artifact_metadata_root(project_name, experiment_id)

        try:
            self._rm(f"{artifact_metadata_root}/{artifact_id}")
        except FileNotFoundError:
            raise RubiconException(f"No artifact with id `{artifact_id}` found.")

    # ------- Dataframes -------

    def _get_dataframe_metadata_root(self, project_name, experiment_id=None):
        """Returns the dataframes directory of the project with name
        `project_name` or experiment with ID `experiment_id`.
        """
        if experiment_id is not None:
            experiment_metadata_root = self._get_experiment_metadata_root(project_name)

            return f"{experiment_metadata_root}/{experiment_id}/dataframes"
        else:
            return f"{self.root_dir}/{slugify(project_name)}/dataframes"

    def _get_dataframe_metadata_path(self, project_name, experiment_id, dataframe_id):
        """Returns the path of the dataframe with ID `dataframe_id`'s
        metadata.
        """
        dataframe_metadata_root = self._get_dataframe_metadata_root(project_name, experiment_id)

        return f"{dataframe_metadata_root}/{dataframe_id}/metadata.json"

    def _get_dataframe_data_path(self, project_name, experiment_id, dataframe_id):
        """Returns the path of the dataframe with ID `dataframe_id`'s
        raw data.
        """
        dataframe_metadata_root = self._get_dataframe_metadata_root(project_name, experiment_id)

        return f"{dataframe_metadata_root}/{dataframe_id}/data"

    def _persist_dataframe(
        self, df: Union[pd.DataFrame, "dd.DataFrame", "pl.DataFrame"], path: str
    ):
        """Persists the dataframe `df` to the configured filesystem.

        Note
        ----
        `dask` dataframes will automatically be split into chunks by `dask.dataframe.to_parquet`.
        `pandas` dataframes, however, will be saved as a single file with the hope that users
        would leverage dask for large dataframes.
        """
        if isinstance(df, pd.DataFrame):
            self._mkdir(path)
            path = f"{path}/data.parquet"

        if hasattr(df, "write_parquet"):
            # handle Polars
            df.write_parquet(path)
        else:
            # Dask or pandas
            df.to_parquet(path, engine="pyarrow", storage_options=self._df_storage_options)

    def _read_dataframe(self, path, df_type: Literal["pandas", "dask", "polars"] = "pandas"):
        """Reads the dataframe `df` from the configured filesystem."""
        df = None
        acceptable_types = ["pandas", "dask", "polars"]

        if df_type == "pandas":
            path = f"{path}/data.parquet"
            df = pd.read_parquet(path, engine="pyarrow", storage_options=self._df_storage_options)
        elif df_type == "polars":
            try:
                from polars import read_parquet
            except ImportError:
                raise RubiconException(
                    "`rubicon_ml` requires `polars` to be installed in the current environment "
                    "to read dataframes with `df_type`='polars'. `pip install polars` "
                    "or `conda install polars` to continue."
                )
            df = read_parquet(path, storage_options=self._df_storage_options)

        elif df_type == "dask":
            try:
                from dask import dataframe as dd
            except ImportError:
                raise RubiconException(
                    "`rubicon_ml` requires `dask` to be installed in the current environment "
                    "to read dataframes with `df_type`='dask'. `pip install dask[dataframe]` "
                    "or `conda install dask` to continue."
                )

            df = dd.read_parquet(path, engine="pyarrow", storage_options=self._df_storage_options)
        else:
            raise ValueError(f"`df_type` must be one of {acceptable_types}")

        return df

    def create_dataframe(
        self,
        dataframe: domain.Dataframe,
        data: Union[pd.DataFrame, "dd.DataFrame", "pl.DataFrame"],
        project_name: str,
        experiment_id: Optional[str] = None,
    ):
        """Persist a dataframe to the configured filesystem.

        Parameters
        ----------
        dataframe : rubicon.domain.Dataframe
            The dataframe to persist.
        data : dask.dataframe.DataFrame or pandas.DataFrame
            The raw data to persist as a dataframe. All
            dataframes are persisted as `dask` dataframes.
        project_name : str
            The name of the project this dataframe belongs to.
        experiment_id : str, optional
            The ID of the experiment this dataframe belongs to.
            Dataframes do not need to belong to an experiment.
        """
        dataframe_metadata_path = self._get_dataframe_metadata_path(
            project_name, experiment_id, dataframe.id
        )
        dataframe_data_path = self._get_dataframe_data_path(
            project_name, experiment_id, dataframe.id
        )

        self._persist_dataframe(data, dataframe_data_path)
        self._persist_domain(dataframe, dataframe_metadata_path)

    def get_dataframe_metadata(self, project_name, dataframe_id, experiment_id=None):
        """Retrieve a dataframes's metadata from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the dataframe with ID
            `dataframe_id` is logged to.
        dataframe_id : str
            The ID of the dataframe to retrieve.
        experiment_id : str, optional
            The ID of the experiment the dataframe with ID
            `dataframe_id` is logged to. Dataframes do not
            need to belong to an experiment.

        Returns
        -------
        rubicon.domain.Dataframe
            The dataframe with ID `dataframe_id`.
        """
        dataframe_metadata_path = self._get_dataframe_metadata_path(
            project_name, experiment_id, dataframe_id
        )
        dataframe = self._read_domain(
            dataframe_metadata_path,
            f"No dataframe with id `{dataframe_id}` found.",
        )

        return domain.Dataframe(**dataframe)

    def get_dataframes_metadata(
        self, project_name: str, experiment_id: Optional[str] = None
    ) -> List[domain.Dataframe]:
        """Retrieve all dataframes' metadata from the configured
        filesystem that belong to the specified object.

        Parameters
        ----------
        project_name : str
            The name of the project to retrieve all dataframes
            from.
        experiment_id : str, optional
            The ID of the experiment to retrieve all dataframes
            from. Dataframes do not need to belong to an
            experiment.

        Returns
        -------
        list of rubicon.domain.Dataframe
            The dataframes logged to the specified object.
        """
        dataframe_metadata_root = self._get_dataframe_metadata_root(project_name, experiment_id)

        return self._load_metadata_files(dataframe_metadata_root, domain.Dataframe)

    def get_dataframe_data(
        self,
        project_name: str,
        dataframe_id: str,
        experiment_id: Optional[str] = None,
        df_type: Literal["pandas", "dask", "polars"] = "pandas",
    ):
        """Retrieve a dataframe's raw data.

        Parameters
        ----------
        project_name : str
            The name of the project the dataframe with ID
            `dataframe_id` is logged to.
        dataframe_id : str
            The ID of the dataframe to retrieve data from.
        experiment_id : str, optional
            The ID of the experiment the dataframe with ID
            `artifact_id` is logged to. Dataframes do not
            need to belong to an experiment.
        df_type : str, optional
            The type of dataframe. Can be `pandas`, `dask`, or `polars`.

        Returns
        -------
        dask.dataframe.DataFrame
            The dataframe with ID `dataframe_id`'s raw data.
        """
        dataframe_data_path = self._get_dataframe_data_path(
            project_name, experiment_id, dataframe_id
        )

        try:
            df = self._read_dataframe(dataframe_data_path, df_type)
        except FileNotFoundError:
            raise RubiconException(
                f"No data for dataframe with id `{dataframe_id}` found. This might have "
                "happened if you forgot to set `df_type='dask'` when trying to read a `dask` dataframe."
            )

        return df

    def delete_dataframe(self, project_name, dataframe_id, experiment_id=None):
        """Delete a dataframe from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the dataframe with ID
            `dataframe_id` is logged to.
        dataframe_id : str
            The ID of the dataframe to delete.
        experiment_id : str, optional
            The ID of the experiment the dataframe with ID
            `artifact_id` is logged to. Dataframes do not
            need to belong to an experiment.
        """
        dataframe_metadata_root = self._get_dataframe_metadata_root(project_name, experiment_id)

        try:
            self._rm(f"{dataframe_metadata_root}/{dataframe_id}")
        except FileNotFoundError:
            raise RubiconException(f"No dataframe with id `{dataframe_id}` found.")

    # -------- Features --------

    def _get_feature_metadata_root(self, project_name, experiment_id):
        """Returns the features directory of the experiment with
        ID `experiment_id`.
        """
        experiment_metadata_root = self._get_experiment_metadata_root(project_name)

        return f"{experiment_metadata_root}/{experiment_id}/features"

    def _get_feature_metadata_path(self, project_name, experiment_id, feature_name):
        """Returns the path of the feature with name `feature_name`'s
        metadata.
        """
        feature_metadata_root = self._get_feature_metadata_root(project_name, experiment_id)

        return f"{feature_metadata_root}/{slugify(feature_name)}/metadata.json"

    def create_feature(self, feature, project_name, experiment_id):
        """Persist a feature to the configured filesystem.

        Parameters
        ----------
        feature : rubicon.domain.Feature
            The feature to persist.
        project_name : str
            The name of the project the experiment with ID
            `experiment_id` is logged to.
        experiment_id : str
            The ID of the experiment this feature belongs to.
        """
        feature_metadata_path = self._get_feature_metadata_path(
            project_name, experiment_id, feature.name
        )

        if self._exists(feature_metadata_path):
            raise RubiconException(f"A feature with name '{feature.name}' already exists.")

        self._persist_domain(feature, feature_metadata_path)

    def get_feature(
        self, project_name: str, experiment_id: str, feature_name: str
    ) -> domain.Feature:
        """Retrieve a feature from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the experiment with ID
            `experiment_id` is logged to.
        experiment_id : str
            The ID of the experiment the feature with name
            `feature_name` is logged to.
        feature_name : str
            The name of the feature to retrieve.

        Returns
        -------
        rubicon.domain.Feature
            The feature with name `feature_name`.
        """
        feature_metadata_path = self._get_feature_metadata_path(
            project_name, experiment_id, feature_name
        )
        feature = self._read_domain(
            feature_metadata_path,
            f"No feature with name '{feature_name}' found.",
        )

        return domain.Feature(**feature)

    def get_features(self, project_name: str, experiment_id: str) -> List[domain.Feature]:
        """Retrieve all features from the configured filesystem
        that belong to the experiment with ID `experiment_id`.

        Parameters
        ----------
        project_name : str
            The name of the project the experiment with ID
            `experiment_id` is logged to.
        experiment_id : str
            The ID of the experiment to retrieve all features
            from.

        Returns
        -------
        list of rubicon.domain.Feature
            The features logged to the experiment with ID
            `experiment_id`.
        """
        feature_metadata_root = self._get_feature_metadata_root(project_name, experiment_id)

        return self._load_metadata_files(feature_metadata_root, domain.Feature)

    # -------- Metrics ---------

    def _get_metric_metadata_root(self, project_name, experiment_id):
        """Returns the metrics directory of the experiment with
        ID `experiment_id`.
        """
        experiment_metadata_root = self._get_experiment_metadata_root(project_name)

        return f"{experiment_metadata_root}/{experiment_id}/metrics"

    def _get_metric_metadata_path(self, project_name, experiment_id, metric_name):
        """Returns the path of the metric with name `metric_name`'s
        metadata.
        """
        metric_metadata_root = self._get_metric_metadata_root(project_name, experiment_id)

        return f"{metric_metadata_root}/{slugify(metric_name)}/metadata.json"

    def create_metric(self, metric, project_name, experiment_id):
        """Persist a metric to the configured filesystem.

        Parameters
        ----------
        metric : rubicon.domain.Metric
            The metric to persist.
        project_name : str
            The name of the project the experiment with ID
            `experiment_id` is logged to.
        experiment_id : str
            The ID of the experiment this metric belongs to.
        """
        metric_metadata_path = self._get_metric_metadata_path(
            project_name, experiment_id, metric.name
        )

        if self._exists(metric_metadata_path):
            raise RubiconException(f"A metric with name '{metric.name}' already exists.")

        self._persist_domain(metric, metric_metadata_path)

    def get_metric(self, project_name: str, experiment_id: str, metric_name: str) -> domain.Metric:
        """Retrieve a metric from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the experiment with ID
            `experiment_id` is logged to.
        experiment_id : str
            The ID of the experiment the metric with name
            `metric_name` is logged to.
        metric_name : str
            The name of the metric to retrieve.

        Returns
        -------
        rubicon.domain.Metric
            The metric with name `metric_name`.
        """
        metric_metadata_path = self._get_metric_metadata_path(
            project_name, experiment_id, metric_name
        )
        metric = self._read_domain(
            metric_metadata_path,
            f"No metric with name '{metric_name}' found.",
        )

        return domain.Metric(**metric)

    def get_metrics(self, project_name: str, experiment_id: str) -> List[domain.Metric]:
        """Retrieve all metrics from the configured filesystem
        that belong to the experiment with ID `experiment_id`.

        Parameters
        ----------
        project_name : str
            The name of the project the experiment with ID
            `experiment_id` is logged to.
        experiment_id : str
            The ID of the experiment to retrieve all metrics
            from.

        Returns
        -------
        list of rubicon.domain.Metric
            The metrics logged to the experiment with ID
            `experiment_id`.
        """
        metric_metadata_root = self._get_metric_metadata_root(project_name, experiment_id)

        return self._load_metadata_files(metric_metadata_root, domain.Metric)

    # ------- Parameters -------

    def _get_parameter_metadata_root(self, project_name, experiment_id):
        """Returns the parameters directory of the experiment with
        ID `experiment_id`.
        """
        experiment_metadata_root = self._get_experiment_metadata_root(project_name)

        return f"{experiment_metadata_root}/{experiment_id}/parameters"

    def _get_parameter_metadata_path(self, project_name, experiment_id, parameter_name):
        """Returns the path of the parameter with name `parameter_name`'s
        metadata.
        """
        parameter_metadata_root = self._get_parameter_metadata_root(project_name, experiment_id)

        return f"{parameter_metadata_root}/{slugify(parameter_name)}/metadata.json"

    def create_parameter(self, parameter, project_name, experiment_id):
        """Persist a parameter to the configured filesystem.

        Parameters
        ----------
        parameter : rubicon.domain.Parameter
            The parameter to persist.
        project_name : str
            The name of the project the experiment with ID
            `experiment_id` is logged to.
        experiment_id : str
            The ID of the experiment this parameter belongs to.
        """
        parameter_metadata_path = self._get_parameter_metadata_path(
            project_name, experiment_id, parameter.name
        )

        if self._exists(parameter_metadata_path):
            raise RubiconException(f"A parameter with name '{parameter.name}' already exists.")

        self._persist_domain(parameter, parameter_metadata_path)

    def get_parameter(self, project_name, experiment_id, parameter_name):
        """Retrieve a parameter from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project this parameter belongs to.
        experiment_id : str
            The ID of the experiment the parameter with name
            `parameter_name` is logged to.
        parameter_name : str
            The name of the parameter to retrieve.

        Returns
        -------
        rubicon.domain.Parameter
            The parameter with name `parameter_name`.
        """
        parameter_metadata_path = self._get_parameter_metadata_path(
            project_name, experiment_id, parameter_name
        )
        parameter = self._read_domain(
            parameter_metadata_path,
            f"No parameter with name '{parameter_name}' found.",
        )

        return domain.Parameter(**parameter)

    def get_parameters(self, project_name: str, experiment_id: str) -> List[domain.Parameter]:
        """Retrieve all parameters from the configured filesystem
        that belong to the experiment with ID `experiment_id`.

        Parameters
        ----------
        project_name : str
            The name of the project the experiment with ID
            `experiment_id` is logged to.
        experiment_id : str
            The ID of the experiment to retrieve all parameters
            from.

        Returns
        -------
        list of rubicon.domain.Parameter
            The parameters logged to the experiment with ID
            `experiment_id`.
        """
        parameter_metadata_root = self._get_parameter_metadata_root(project_name, experiment_id)

        return self._load_metadata_files(parameter_metadata_root, domain.Parameter)

    # ---------- Tags ----------

    def _get_tag_metadata_root(
        self, project_name, experiment_id=None, entity_identifier=None, entity_type=None
    ):
        """Returns the directory to write tags to."""
        get_metadata_root_lookup = {
            "Artifact": self._get_artifact_metadata_root,
            "Dataframe": self._get_dataframe_metadata_root,
            "Experiment": self._get_experiment_metadata_root,
            "Metric": self._get_metric_metadata_root,
            "Feature": self._get_feature_metadata_root,
            "Parameter": self._get_parameter_metadata_root,
        }

        try:
            get_metadata_root = get_metadata_root_lookup[entity_type]
        except KeyError:
            raise ValueError("`experiment_id` and `entity_identifier` can not both be `None`.")

        if entity_type == "Experiment":
            experiment_metadata_root = get_metadata_root(project_name)

            return f"{experiment_metadata_root}/{experiment_id}"
        else:
            entity_metadata_root = get_metadata_root(project_name, experiment_id)

            # We want to slugify the names of Metrics, Features, and Parameters- not Artifacts, Dataframes, or Experiments
            if entity_type in ["Metric", "Feature", "Parameter"]:
                entity_identifier = slugify(entity_identifier)
            return f"{entity_metadata_root}/{entity_identifier}"

    def add_tags(
        self,
        project_name,
        tags,
        experiment_id=None,
        entity_identifier=None,
        entity_type=None,
    ):
        """Persist tags to the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the object to tag
            belongs to.
        tags : list of str
            The tag values to persist.
        experiment_id : str, optional
            The ID of the experiment to apply the tags
            `tags` to.
        entity_identifier : str, optional
            The ID or name of the entity to apply the tags
            `tags` to.
        entity_type : str, optional
            The name of the entity's type as returned by
            `entity_cls.__class__.__name__`.
        """
        tag_metadata_root = self._get_tag_metadata_root(
            project_name, experiment_id, entity_identifier, entity_type
        )
        tag_metadata_path = f"{tag_metadata_root}/tags_{uuid4()}.json"

        self._persist_domain({"added_tags": tags}, tag_metadata_path)

    def remove_tags(
        self,
        project_name,
        tags,
        experiment_id=None,
        entity_identifier=None,
        entity_type=None,
    ):
        """Delete tags from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the object to delete
            tags from belongs to.
        tags : list of str
            The tag values to delete.
        experiment_id : str, optional
            The ID of the experiment to delete the tags
            `tags` from.
        entity_identifier : str, optional
            The ID or name of the entity to apply the tags
            `tags` to.
        entity_type : str, optional
            The name of the entity's type as returned by
            `entity_cls.__class__.__name__`.
        """
        tag_metadata_root = self._get_tag_metadata_root(
            project_name, experiment_id, entity_identifier, entity_type
        )
        tag_metadata_path = f"{tag_metadata_root}/tags_{uuid4()}.json"

        self._persist_domain({"removed_tags": tags}, tag_metadata_path)

    def _sort_tag_paths(self, tag_paths):
        """Sorts the paths in `tags_paths` by when they were
        created.
        """
        if isinstance(tag_paths, dict):
            tag_paths = tag_paths.values()

        tag_paths_with_timestamps = [
            (t.get("created", t.get("LastModified")), t.get("name")) for t in tag_paths
        ]
        tag_paths_with_timestamps.sort()

        return tag_paths_with_timestamps

    def get_tags(self, project_name, experiment_id=None, entity_identifier=None, entity_type=None):
        """Retrieve tags from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the object to retrieve
            tags from belongs to.
        experiment_id : str, optional
            The ID of the experiment to retrieve tags from.
        entity_identifier : str, optional
            The ID or name of the entity to apply the tags
            `tags` to.
        entity_type : str, optional
            The name of the entity's type as returned by
            `entity_cls.__class__.__name__`.

        Returns
        -------
        list of dict
            A list of dictionaries with one key each,
            `added_tags` or `removed_tags`, where the
            value is a list of tag names that have been
            added to or removed from the specified object.
        """
        tag_metadata_root = self._get_tag_metadata_root(
            project_name, experiment_id, entity_identifier, entity_type
        )
        tag_metadata_glob = f"{tag_metadata_root}/tags_*.json"

        tag_paths = self._glob(tag_metadata_glob)
        if len(tag_paths) == 0:
            return []

        sorted_tag_paths = self._sort_tag_paths(tag_paths)

        tag_data = self._cat([p for _, p in sorted_tag_paths])
        sorted_tag_data = [json.loads(tag_data[p]) for _, p in sorted_tag_paths]

        return sorted_tag_data

    # ---------- Comments ----------

    def _get_comment_metadata_root(
        self, project_name, experiment_id=None, entity_identifier=None, entity_type=None
    ):
        """Returns the directory to write comments to."""
        # comments and tags are currently written to the same root with a different filename
        return self._get_tag_metadata_root(
            project_name, experiment_id, entity_identifier, entity_type
        )

    def add_comments(
        self,
        project_name,
        comments,
        experiment_id=None,
        entity_identifier=None,
        entity_type=None,
    ):
        """Persist comments to the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the object to comment
            belongs to.
        comments : list of str
            The comment values to persist.
        experiment_id : str, optional
            The ID of the experiment to apply the comments
            `comments` to.
        entity_identifier : str, optional
            The ID or name of the entity to apply the comments
            `comments` to.
        entity_type : str, optional
            The name of the entity's type as returned by
            `entity_cls.__class__.__name__`.
        """
        comment_metadata_root = self._get_comment_metadata_root(
            project_name, experiment_id, entity_identifier, entity_type
        )
        comment_metadata_path = f"{comment_metadata_root}/comments_{uuid4()}.json"

        self._persist_domain({"added_comments": comments}, comment_metadata_path)

    def remove_comments(
        self,
        project_name,
        comments,
        experiment_id=None,
        entity_identifier=None,
        entity_type=None,
    ):
        """Delete comments from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the object to delete
            comments from belongs to.
        comments : list of str
            The comment values to delete.
        experiment_id : str, optional
            The ID of the experiment to delete the comments
            `comments` from.
        entity_identifier : str, optional
            The ID or name of the entity to apply the comments
            `comments` to.
        entity_type : str, optional
            The name of the entity's type as returned by
            `entity_cls.__class__.__name__`.
        """
        comment_metadata_root = self._get_comment_metadata_root(
            project_name, experiment_id, entity_identifier, entity_type
        )
        comment_metadata_path = f"{comment_metadata_root}/comments_{uuid4()}.json"

        self._persist_domain({"removed_comments": comments}, comment_metadata_path)

    def _sort_comment_paths(self, comment_paths):
        """Sorts the paths in `comment_paths` by when they were
        created.
        """
        return self._sort_tag_paths(comment_paths)

    def get_comments(
        self, project_name, experiment_id=None, entity_identifier=None, entity_type=None
    ):
        """Retrieve comments from the configured filesystem.

        Parameters
        ----------
        project_name : str
            The name of the project the object to retrieve
            comments from belongs to.
        experiment_id : str, optional
            The ID of the experiment to retrieve comments from.
        entity_identifier : str, optional
            The ID or name of the entity to apply the comments
            `comments` to.
        entity_type : str, optional
            The name of the entity's type as returned by
            `entity_cls.__class__.__name__`.

        Returns
        -------
        dict
            A dictionary, `added_comments` where the
            value is a list of comment names that have
            been added to the specified object.
        """
        comment_metadata_root = self._get_comment_metadata_root(
            project_name, experiment_id, entity_identifier, entity_type
        )
        comment_metadata_glob = f"{comment_metadata_root}/comments_*.json"

        comment_paths = self._glob(comment_metadata_glob)
        if len(comment_paths) == 0:
            return []

        sorted_comment_paths = self._sort_comment_paths(comment_paths)

        comment_data = self._cat([p for _, p in sorted_comment_paths])
        sorted_comment_data = [json.loads(comment_data[p]) for _, p in sorted_comment_paths]

        return sorted_comment_data



from __future__ import annotations

import subprocess
import warnings
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

import pandas as pd

from rubicon_ml import domain
from rubicon_ml.client import ArtifactMixin, Base, DataframeMixin, Experiment
from rubicon_ml.client.utils.exception_handling import failsafe
from rubicon_ml.client.utils.tags import filter_children
from rubicon_ml.exceptions import RubiconException
from rubicon_ml.schema.logger import SchemaMixin

if TYPE_CHECKING:
    import dask.dataframe as dd

    from rubicon_ml import Rubicon
    from rubicon_ml.client import Config, Dataframe
    from rubicon_ml.domain import Project as ProjectDomain


class Project(Base, ArtifactMixin, DataframeMixin, SchemaMixin):
    """A client project.

    A `project` is a collection of `experiments`,
    `dataframes`, and `artifacts` identified by a unique name.

    Parameters
    ----------
    domain : rubicon.domain.Project
        The project domain model.
    config : rubicon.client.Config
        The config, which specifies the underlying repository.
    """

    def __init__(self, domain: ProjectDomain, config: Optional[Union[Config, List[Config]]] = None):
        super().__init__(domain, config)

        self._domain: ProjectDomain

        self._artifacts = []
        self._dataframes = []
        self._experiments = []

    def _get_branch_name(self) -> str:
        """Returns the name of the active branch of the `git` repo
        it is called from.
        """
        command = ["git", "rev-parse", "--abbrev-ref", "HEAD"]
        completed_process = subprocess.run(command, capture_output=True)

        return completed_process.stdout.decode("utf8").replace("\n", "")

    def _get_commit_hash(self) -> str:
        """Returns the hash of the last commit to the active branch
        of the `git` repo it is called from.
        """
        command = ["git", "rev-parse", "HEAD"]
        completed_process = subprocess.run(command, capture_output=True)

        return completed_process.stdout.decode("utf8").replace("\n", "")

    def _get_identifiers(self) -> Tuple[Optional[str], None]:
        """Get the project's name."""
        return self.name, None

    def _create_experiment_domain(
        self,
        name,
        description,
        model_name,
        branch_name,
        commit_hash,
        training_metadata,
        tags,
        comments,
    ):
        """Instantiates and returns an experiment domain object."""
        if self.is_auto_git_enabled():
            if branch_name is None:
                branch_name = self._get_branch_name()
            if commit_hash is None:
                commit_hash = self._get_commit_hash()

        if training_metadata is not None:
            training_metadata = domain.utils.TrainingMetadata(training_metadata)

        return domain.Experiment(
            project_name=self._domain.name,
            name=name,
            description=description,
            model_name=model_name,
            branch_name=branch_name,
            commit_hash=commit_hash,
            training_metadata=training_metadata,
            tags=tags,
            comments=comments,
        )

    def _group_experiments(self, experiments: List[Experiment], group_by: Optional[str] = None):
        """Groups experiments by `group_by`. Valid options include ["commit_hash"].

        Returns
        -------
        dict
            A dictionary of (group name, DataFrame) key-value pairs.
        """
        GROUP_BY_OPTIONS = ["commit_hash"]
        if group_by is not None and group_by not in GROUP_BY_OPTIONS:
            raise ValueError(f"`group_by` must be one of {GROUP_BY_OPTIONS} or `None`.")

        if group_by is not None:
            grouped_experiments = {}

            if group_by == "commit_hash":
                for experiment in experiments:
                    current_experiments = grouped_experiments.get(experiment.commit_hash, [])
                    current_experiments.append(experiment)
                    grouped_experiments[experiment.commit_hash] = current_experiments
        else:
            grouped_experiments = {None: experiments}

        return grouped_experiments

    def to_dask_df(self, group_by: Optional[str] = None):
        """DEPRECATED: Available for backwards compatibility."""
        warnings.warn(
            "`to_dask_df` is deprecated and will be removed in a future release. "
            "use `to_df(df_type='dask') instead.",
            DeprecationWarning,
        )

        return self.to_df(df_type="dask", group_by=group_by)

    @failsafe
    def to_df(
        self, df_type: str = "pandas", group_by: Optional[str] = None
    ) -> Union[pd.DataFrame, Dict[str, pd.DataFrame], dd.DataFrame, Dict[str, dd.DataFrame]]:
        """Loads the project's data into dask or pandas dataframe(s) sorted by
        `created_at`. This includes the experiment details along with parameters
        and metrics.

        Parameters
        ----------
        df_type : str, optional
            The type of dataframe to return. Valid options include
            ["dask", "pandas"]. Defaults to "pandas".
        group_by : str or None, optional
            How to group the project's experiments in the returned
            dataframe(s). Valid options include ["commit_hash"].

        Returns
        -------
        pandas.DataFrame or dict of pandas.DataFrame or dask.DataFrame or dict of dask.DataFrame
            If `group_by` is `None`, a dask or pandas dataframe holding the project's
            data. Otherwise a dict of dask or pandas dataframes holding the project's
            data grouped by `group_by`.
        """
        DEFAULT_COLUMNS = [
            "id",
            "name",
            "description",
            "model_name",
            "commit_hash",
            "tags",
            "created_at",
        ]

        experiments = self.experiments()
        grouped_experiments = self._group_experiments(experiments, group_by=group_by)

        experiment_dfs = {}
        for group, experiments in grouped_experiments.items():
            experiment_records = []
            parameter_names = set()
            metric_names = set()

            for experiment in experiments:
                experiment_record = {
                    "id": experiment.id,
                    "name": experiment.name,
                    "description": experiment.description,
                    "model_name": experiment.model_name,
                    "commit_hash": experiment.commit_hash,
                    "tags": experiment.tags,
                    "created_at": experiment.created_at,
                }

                for parameter in experiment.parameters():
                    experiment_record[f"{parameter.name}"] = parameter.value
                    parameter_names.add(parameter.name)

                for metric in experiment.metrics():
                    experiment_record[f"{metric.name}"] = metric.value
                    metric_names.add(metric.name)

                # TODO - features, artifacts, dataframes represented here?

                experiment_records.append(experiment_record)

            columns = DEFAULT_COLUMNS + list(parameter_names) + list(metric_names)
            df = pd.DataFrame.from_records(experiment_records, columns=columns)
            df = df.sort_values(by=["created_at"], ascending=False).reset_index(drop=True)

            if df_type == "dask":
                try:
                    from dask import dataframe as dd
                except ImportError:
                    raise RubiconException(
                        "`rubicon_ml` requires `dask` to be installed in the current "
                        "environment to create dataframes with `df_type`='dask'. `pip install "
                        "dask[dataframe]` or `conda install dask` to continue."
                    )

                df = dd.from_pandas(df, npartitions=1)

            experiment_dfs[group] = df

        return experiment_dfs if group_by is not None else list(experiment_dfs.values())[0]

    @failsafe
    def log_experiment(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        model_name: Optional[str] = None,
        branch_name: Optional[str] = None,
        commit_hash: Optional[str] = None,
        training_metadata: Optional[Union[Tuple, List[Tuple]]] = None,
        tags: Optional[List[str]] = None,
        comments: Optional[List[str]] = None,
    ) -> Experiment:
        """Log a new experiment to this project.

        Parameters
        ----------
        name : str
            The experiment's name.
        description : str, optional
            The experiment's description. Use to provide
            additional context.
        model_name : str, optional
            The experiment's model name. For example, this
            could be the name of the registered model in Model One.
        branch_name : str, optional
            The name of the active branch of the `git` repo this experiment
            is logged from. If omitted and automatic `git` logging is enabled,
            it will be retrieved via `git rev-parse`.
        commit_hash : str, optional
            The hash of the last commit to the active branch of the `git` repo
            this experiment is logged from. If omitted and automatic `git`
            logging is enabled, it will be retrieved via `git rev-parse`.
        training_metadata : tuple or list of tuples, optional
            Metadata associated with the experiment's
            training dataset(s).
        tags : list of str, optional
            Values to tag the experiment with. Use tags to organize and
            filter your experiments. For example, tags could be used
            to differentiate between the type of model or classifier
            used during the experiment (i.e. `linear regression`
            or `random forest`).
        comments : list of str, optional
            Values to comment the experiment with.

        Returns
        -------
        rubicon.client.Experiment
            The created experiment.
        """
        if tags is None:
            tags = []

        if not isinstance(tags, list) or not all([isinstance(tag, str) for tag in tags]):
            raise ValueError("`tags` must be `list` of type `str`")

        if comments is None:
            comments = []

        if not isinstance(comments, list) or not all(
            [isinstance(comment, str) for comment in comments]
        ):
            raise ValueError("`comments` must be `list` of type `str`")

        experiment = self._create_experiment_domain(
            name,
            description,
            model_name,
            branch_name,
            commit_hash,
            training_metadata,
            tags,
            comments,
        )
        for repo in self.repositories:
            repo.create_experiment(experiment)

        return Experiment(experiment, self)

    @failsafe
    def experiment(self, id: Optional[str] = None, name: Optional[str] = None) -> Experiment:
        """Get an experiment logged to this project by id or name.

        Parameters
        ----------
        id : str
            The id of the experiment to get.
        name : str
            The name of the experiment to get.

        Returns
        -------
        rubicon.client.Experiment
            The experiment logged to this project with id `id` or name 'name'.
        """
        if (name is None and id is None) or (name is not None and id is not None):
            raise ValueError("`name` OR `id` required.")

        if name is not None:
            experiments = [e for e in self.experiments() if e.name == name]

            if len(experiments) == 0:
                raise RubiconException(f"No experiment found with name '{name}'.")
            elif len(experiments) > 1:
                warnings.warn(
                    f"Multiple experiments found with name '{name}'."
                    " Returning most recently logged."
                )

            return experiments[-1]
        else:
            return_err = None

            for repo in self.repositories:
                try:
                    experiment = Experiment(repo.get_experiment(self.name, id), self)
                except Exception as err:
                    return_err = err
                else:
                    return experiment

            self._raise_rubicon_exception(return_err)

    @failsafe
    def experiments(
        self, tags: Optional[List[str]] = None, qtype: str = "or", name: Optional[str] = None
    ) -> List[Experiment]:
        """Get the experiments logged to this project.

        Parameters
        ----------
        tags : list of str, optional
            The tag values to filter results on.
        qtype : str, optional
            The query type to filter results on. Can be 'or' or
            'and'. Defaults to 'or'.
        name:
            The name of the experiment(s) to filter results on.

        Returns
        -------
        list of rubicon.client.Experiment
            The experiments previously logged to this project.
        """
        if tags is None:
            tags = []

        return_err = None

        for repo in self.repositories:
            try:
                experiments = [Experiment(e, self) for e in repo.get_experiments(self.name)]
            except Exception as err:
                return_err = err
            else:
                self._experiments = filter_children(experiments, tags, qtype, name)

                return self._experiments

        self._raise_rubicon_exception(return_err)

    @failsafe
    def dataframes(
        self,
        tags: Optional[List[str]] = None,
        qtype: str = "or",
        recursive: bool = False,
        name: Optional[str] = None,
    ) -> List[Dataframe]:
        """Get the dataframes logged to this project.

        Parameters
        ----------
        tags : list of str, optional
            The tag values to filter results on.
        qtype : str, optional
            The query type to filter results on. Can be 'or' or
            'and'. Defaults to 'or'.
        recursive : bool, optional
            If true, get the dataframes logged to this project's
            experiments as well. Defaults to false.
        name : str
            The name value to filter results on.

        Returns
        -------
        list of rubicon.client.Dataframe
            The dataframes previously logged to this client object.
        """
        if tags is None:
            tags = []
        super().dataframes(tags=tags, qtype=qtype, name=name)

        if recursive is True:
            for experiment in self.experiments():
                self._dataframes.extend(experiment.dataframes(tags=tags, qtype=qtype, name=name))

        return self._dataframes

    @failsafe
    def archive(
        self,
        experiments: Optional[List[Experiment]] = None,
        remote_rubicon: Optional[Rubicon] = None,
    ):
        """Archive the experiments logged to this project.

        Parameters
        ----------
        experiments : list of Experiments, optional
            The rubicon.client.Experiment objects to archive. If None all logged experiments are archived.
        remote_rubicon : rubicon_ml.Rubicon object, optional
            The remote Rubicon object with the repository to archive to

        Returns
        -------
        filepath of newly created archive
        """
        if len(self.experiments()) == 0:
            raise ValueError("`project` has no logged `experiments` to archive")
        if experiments is not None:
            if not isinstance(experiments, list) or not all(
                [isinstance(experiment, Experiment) for experiment in experiments]
            ):
                raise ValueError(
                    "`experiments` must be `list` of type `rubicon_ml.client.Experiment`"
                )

        if remote_rubicon is not None:
            from rubicon_ml import Rubicon

            if not isinstance(remote_rubicon, Rubicon):
                raise ValueError("`remote_rubicon` must be of type `rubicon_ml.client.Rubicon`")
            else:
                return self.repository._archive(
                    self.name, experiments, remote_rubicon.repository.root_dir
                )
        else:
            return self.repository._archive(self.name, experiments, None)

    @failsafe
    def experiments_from_archive(self, remote_rubicon, latest_only: Optional[bool] = False):
        """Retrieve archived experiments into this project's experiments folder.

        Parameters
        ----------
        remote_rubicon : rubicon_ml.Rubicon object
            The remote Rubicon object with the repository containing archived experiments to read in
        latest_only : bool, optional
            Indicates whether or not experiments should only be read from the latest archive
        """
        from rubicon_ml import Rubicon

        if not isinstance(remote_rubicon, Rubicon):
            raise ValueError("`remote_rubicon` must be of type `rubicon_ml.client.Rubicon`")
        self.repository._experiments_from_archive(
            self.name, remote_rubicon.repository.root_dir, latest_only
        )

    @property
    def name(self):
        """Get the project's name."""
        return self._domain.name

    @property
    def id(self):
        """Get the project's id."""
        return self._domain.id

    @property
    def description(self):
        """Get the project's description."""
        return self._domain.description

    @property
    def github_url(self):
        """Get the project's GitHub repository URL."""
        return self._domain.github_url

    @property
    def training_metadata(self):
        """Get the project's training metadata."""
        training_metadata = self._domain.training_metadata.training_metadata

        if len(training_metadata) == 1:
            training_metadata = training_metadata[0]

        return training_metadata

    @property
    def created_at(self):
        """Get the time the project was created."""
        return self._domain.created_at



import warnings
from typing import Dict, Literal, Optional, Union

import dash_bootstrap_components as dbc
from dash import Dash, html

from rubicon_ml import __version__ as rubicon_ml_version

_next_available_port = 8050


class VizBase:
    """The base class for all `rubicon_ml` visualizations.

    `VizBase` can not be directly instantatied. New widgets must all
    extend `VizBase`.
    """

    def __init__(
        self,
        dash_title: str = "base",
    ):
        self.dash_title = f"rubicon-ml: {dash_title}"

    @property
    def layout(self):
        raise NotImplementedError("extensions of `VizBase` must implement property `layout(self)`")

    def build_layout(self):
        """Wraps the layout defined by `self.layout` in a container providing
        the `rubicon_ml` header.
        """
        self.app.layout = dbc.Card(
            dbc.CardBody(
                [
                    dbc.Row(
                        [
                            html.Img(
                                id="rubicon-logo-img",
                                src=self.app.get_asset_url("images/rubicon-logo-dark.png"),
                            ),
                        ],
                    ),
                    dbc.Row(html.P(rubicon_ml_version, id="version-text"), id="version-row"),
                    dbc.Row(self.layout),
                ],
                id="frame",
            ),
        )

    def load_experiment_data(self):
        raise NotImplementedError(
            "extensions of `VizBase` must implement `load_experiment_data(self)`"
        )

    def register_callbacks(self, link_experiment_table: bool = False):
        raise NotImplementedError(
            "extensions of `VizBase` must implement `register_callbacks(self)`"
        )

    def serve(
        self,
        in_background: bool = False,
        jupyter_mode: Literal["external", "inline", "jupyterlab", "tab"] = "external",
        dash_kwargs: Dict = {},
        run_server_kwargs: Dict = {},
    ):
        """Serve the Dash app on the next available port to render the visualization.

        Parameters
        ----------
        in_background : bool, optional
            DEPRECATED. Background processing is now handled by `jupyter_mode`.
        jupyter_mode : "external", "inline", "jupyterlab", or "tab", optional
            How to render the dashboard when running from Jupyterlab.
            * "external" to serve the dashboard at an external link.
            * "inline" to render the dashboard in the current notebook's output
              cell.
            * "jupyterlab" to render the dashboard in a new window within the
              current Jupyterlab session.
            * "tab" to serve the dashboard at an external link and open a new
              browser tab to said link.
            Defaults to "external".
        dash_kwargs : dict, optional
            Keyword arguments to be passed along to the newly instantiated
            Dash object. Available options can be found at
            https://dash.plotly.com/reference#dash.dash.
        run_server_kwargs : dict, optional
            Keyword arguments to be passed along to `Dash.run_server`.
            Available options can be found at
            https://dash.plotly.com/reference#app.run_server. Most commonly,
            the 'port' argument can be provided here to serve the app on a
            specific port.
        """
        if in_background:
            warnings.warn(
                "The `in_background` argument is deprecated and will have no effect, "
                "Background processing is now handled by `jupyter_mode`.",
                DeprecationWarning,
            )

        JUPYTER_MODES = ["external", "inline", "jupyterlab", "tab"]
        if jupyter_mode not in JUPYTER_MODES:
            raise ValueError(
                f"Invalid `jupyter_mode` '{jupyter_mode}'. Must be one of "
                f"{', '.join(JUPYTER_MODES)}"
            )

        if self.experiments is None:
            raise RuntimeError(
                f"`{self.__class__}.experiments` can not be None when `serve` is called"
            )

        self.app = Dash(
            __name__,
            external_stylesheets=[dbc.themes.LUX, dbc.icons.BOOTSTRAP],
            title=self.dash_title,
            **dash_kwargs,
        )

        self.load_experiment_data()
        self.build_layout()
        self.register_callbacks()

        global _next_available_port

        default_run_server_kwargs = {
            "dev_tools_silence_routes_logging": True,
            "port": _next_available_port,
        }
        default_run_server_kwargs.update(run_server_kwargs)

        if jupyter_mode != "inline":
            default_run_server_kwargs["jupyter_mode"] = jupyter_mode

        _next_available_port = default_run_server_kwargs["port"] + 1

        self.app.run(**default_run_server_kwargs)

    def show(
        self,
        i_frame_kwargs: Dict = {},
        dash_kwargs: Dict = {},
        run_server_kwargs: Dict = {},
        height: Optional[Union[int, str]] = None,
        width: Optional[Union[int, str]] = None,
    ):
        """Serve the Dash app on the next available port to render the visualization.

        Additionally, renders the visualization inline in the current Jupyter notebook.

        Parameters
        ----------
        i_frame_kwargs: dict, optional
            DEPRECATED. Use `height` and `width` instead.
        dash_kwargs : dict, optional
            Keyword arguments to be passed along to the newly instantiated
            Dash object. Available options can be found at
            https://dash.plotly.com/reference#dash.dash.
        run_server_kwargs : dict, optional
            Keyword arguments to be passed along to `Dash.run_server`.
            Available options can be found at
            https://dash.plotly.com/reference#app.run_server. Most commonly,
            the 'port' argument can be provided here to serve the app on a
            specific port.
        height : int, str or None, optional
            The height of the inline visualizaiton. Integers represent number
            of pixels, strings represent a percentage of the window and must
            end with '%'.
        width : int, str or None, optional
            The width of the inline visualizaiton. Integers represent number
            of pixels, strings represent a percentage of the window and must
            end with '%'.
        """
        if i_frame_kwargs:
            warnings.warn(
                "The `i_frame_kwargs` argument is deprecated and will have no effect, "
                "use `height` and `width` instead.",
                DeprecationWarning,
            )

        if height is not None:
            run_server_kwargs["jupyter_height"] = height
        if width is not None:
            run_server_kwargs["jupyter_width"] = width

        self.serve(
            jupyter_mode="inline",
            dash_kwargs=dash_kwargs,
            run_server_kwargs=run_server_kwargs,
        )



__version__ = "0.12.0"

from rubicon_ml.client import (  # noqa: E402
    Artifact,
    Dataframe,
    Experiment,
    Feature,
    Metric,
    Parameter,
    Project,
    Rubicon,
    RubiconJSON,
)
from rubicon_ml.client.utils.exception_handling import set_failure_mode  # noqa: E402
from rubicon_ml.intake_rubicon.publish import publish  # noqa: E402

__all__ = [
    "Artifact",
    "Dataframe",
    "Experiment",
    "Feature",
    "Metric",
    "Parameter",
    "Project",
    "publish",
    "Rubicon",
    "RubiconJSON",
    "set_failure_mode",
]



import datetime
import logging
import uuid
from typing import Dict, List, Optional

LOGGER = logging.getLogger()


class CommentMixin:
    """Adds comment support to a domain model."""

    def add_comments(self, comments: List[str]):
        """
        Add new comments to this model.

        Parameters
        ----------
        comments : List[str]
            A list of string comments to add to the domain model.
        """
        self.comments.extend(comments)

    def remove_comments(self, comments: List[str]):
        """
        Remove comments from this model.

        Parameters
        ----------
        comments : List[str]
            A list of string comments to remove from this domain model.
        """
        self.comments = list(set(self.comments).difference(set(comments)))


class InitMixin:
    """Adds helpers for initializing default values."""

    def _check_extra_kwargs(self, kwargs: Dict):
        """Warn if extra kwargs are given when initialized.

        Replaces default `dataclass` behavior of erroring on unexpected kwargs.
        """
        if kwargs:
            LOGGER.warning(
                f"{self.__class__.__name__}.__init__() got an unexpected keyword "
                f"argument(s): `{'`, `'.join([key for key in kwargs])}`"
            )

    def _init_created_at(self, created_at: Optional[datetime.datetime]) -> datetime.datetime:
        """Initialize the `created_at` attribute to the current time if necessary.

        `datetime.UTC` was added and `datetime.utcnow` was deprecated in Python 3.11.
        """
        if created_at is None:
            try:
                return datetime.datetime.now(datetime.UTC)
            except AttributeError:
                return datetime.datetime.utcnow()

        return created_at

    def _init_id(self, id: Optional[str]) -> str:
        """Initialize the `id` attribute to a new UUID if necessary."""
        if id is None:
            return str(uuid.uuid4())

        return id


class TagMixin:
    """Adds tagging support to a domain model."""

    def add_tags(self, tags: List[str]):
        """
        Add new tags to this model.

        Parameters
        ----------
        tags : List[str]
            A list of string tags to add to the domain model.
        """
        self.tags = list(set(self.tags).union(set(tags)))

    def remove_tags(self, tags: List[str]):
        """
        Remove tags from this model.

        Parameters
        ----------
        tags : List[str]
            A list of string tags to remove from this domain model.
        """
        self.tags = list(set(self.tags).difference(set(tags)))



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Optional

from rubicon_ml.client import Base, CommentMixin, TagMixin

if TYPE_CHECKING:
    from rubicon_ml.client import Experiment
    from rubicon_ml.domain import Feature as FeatureDomain


class Feature(Base, TagMixin, CommentMixin):
    """A client feature.

    A `feature` is an input to an `experiment` (model run)
    that's an independent, measurable property of a phenomenon
    being observed. It affects the model's predictions.

    For example, consider a model that predicts how likely a
    customer is to pay back a loan. Possible features could be
    'year', 'credit score', etc.

    A `feature` is logged to an `experiment`.

    Parameters
    ----------
    domain : rubicon.domain.Feature
        The feature domain model.
    config : rubicon.client.Config
        The config, which specifies the underlying repository.
    parent : rubicon.client.Experiment
        The experiment that the feature is
        logged to.
    """

    def __init__(self, domain: FeatureDomain, parent: Experiment):
        super().__init__(domain, parent._config)

        self._domain: FeatureDomain

        self._data = None
        self._parent = parent

    @property
    def id(self) -> str:
        """Get the feature's id."""
        return self._domain.id

    @property
    def name(self) -> Optional[str]:
        """Get the feature's name."""
        return self._domain.name

    @property
    def description(self) -> Optional[str]:
        """Get the feature's description."""
        return self._domain.description

    @property
    def importance(self):
        """Get the feature's importance."""
        return self._domain.importance

    @property
    def created_at(self) -> datetime:
        """Get the feature's created_at."""
        return self._domain.created_at

    @property
    def parent(self) -> Experiment:
        """Get the feature's parent client object."""
        return self._parent



import intake  # noqa F401

from rubicon_ml.intake_rubicon.experiment import ExperimentSource
from rubicon_ml.intake_rubicon.viz import (
    DataframePlotDataSource,
    ExperimentsTableDataSource,
    MetricCorrelationPlotDataSource,
    MetricListComparisonDataSource,
)

__all__ = [
    "ExperimentSource",
    "ExperimentsTableDataSource",
    "MetricCorrelationPlotDataSource",
    "DataframePlotDataSource",
    "MetricListComparisonDataSource",
]



import re
from typing import List, Optional, Union


class TagContainer(list):
    """List-based container for tags.

    Allows indexing into tags with colons in them by string, like a dictionary.
    """

    def __getitem__(self, index_or_key: Union[int, str]):
        if isinstance(index_or_key, str):
            values = []

            for tag in self:
                key_value = tag.split(":", 1)

                if len(key_value) > 1 and key_value[0] == index_or_key:
                    values.append(key_value[1])

            if len(values) == 0:
                raise KeyError(index_or_key)
            elif len(values) == 1:
                return values[0]
            else:
                return values
        else:
            item = super().__getitem__(index_or_key)

            return TagContainer(item) if isinstance(item, list) else item


def has_tag_requirements(tags: List[str], required_tags: List[str], qtype: str) -> bool:
    """Determines if the `required_tags` are in `tags`.

    Returns True if all `required_tags` are in `tags` and `qtype` is "and" or if
    any `required_tags` are in `tags` and `qtype` is "or". The tags in
    `required_tags` may contain contain wildcard (*) characters.
    """
    qtype_func = any if qtype == "or" else all

    if any(["*" in tag for tag in required_tags]):

        def _wildcard_match(pattern, tag):
            return re.match(f"^{pattern.replace('*', '.*')}$", tag) is not None

        return qtype_func(
            [
                any([_wildcard_match(required_tag, tag) for tag in tags])
                for required_tag in required_tags
            ]
        )
    else:
        return qtype_func(tag in tags for tag in required_tags)


def filter_children(children, tags: List[str], qtype: str, name: Optional[str]):
    """Return the children in `children` with the given tags or name.

    If both are provided, children are first filtered by tags and then by names.
    """
    if len(tags) > 0:
        children = [c for c in children if has_tag_requirements(c.tags, tags, qtype)]

    if name is not None:
        children = [c for c in children if c.name == name]

    return children



from __future__ import annotations

from typing import TypeVar, Union

from rubicon_ml.domain import utils
from rubicon_ml.domain.artifact import Artifact
from rubicon_ml.domain.dataframe import Dataframe
from rubicon_ml.domain.experiment import Experiment
from rubicon_ml.domain.feature import Feature
from rubicon_ml.domain.metric import Metric
from rubicon_ml.domain.parameter import Parameter
from rubicon_ml.domain.project import Project

DOMAIN_TYPES = Union[Artifact, Dataframe, Experiment, Feature, Metric, Parameter, Project]

DomainsVar = TypeVar(
    "DomainsVar", Artifact, Dataframe, Experiment, Feature, Metric, Parameter, Project
)


__all__ = [
    "Artifact",
    "Dataframe",
    "Experiment",
    "Feature",
    "Metric",
    "Parameter",
    "Project",
    "utils",
]
